import jax
import jax.numpy as jnp
import numpy as np
import astropy
from astropy.cosmology import Planck15
from ..generic import SampledPopulationModel, RedshiftPopulationModel
from ..utils import box

class Redshift(SampledPopulationModel, RedshiftPopulationModel):
    """A class representing redshifts and their associated probability distributions.

    Attributes:
        z_max (float): Maximum redshift.
        lamb_min (float): Minimum value for the hyperparameter.
        lamb_max (float): Maximum value for the hyperparameter.
        var_name (str): Name of the redshift variable.
        hyper_var_name (str): Name of the hyperparameter variable.
        lambs (numpy.ndarray): Array of hyperparameter values.
        zs (numpy.ndarray): Array of redshift values.
        dVdz_values (numpy.ndarray): Array of differential comoving volume values.

    Methods:
        dVdz(data): Computes the differential comoving volume for given redshift data.
        normalization(params): Computes the normalization factor for the probability distribution.
        probability(data, params): Computes the probability distribution for given data and parameters.
        __call__(data, params): Calls the probability method.
    """
    def __init__(self, var_names=['redshift'], hyper_var_names=['lamb'], z_max=3, lamb_min=-10, lamb_max=10):
        """Initialize the Redshift class.

        Args:
            z_max (float, optional): Maximum redshift (default is 3).
            lamb_min (float, optional): Minimum value for the hyperparameter (default is -10).
            lamb_max (float, optional): Maximum value for the hyperparameter (default is 10).
        """
        self.z_max = z_max
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.var_name = var_names[0]
        self.hyper_var_name = hyper_var_names[0]
        self.lamb_min = lamb_min
        self.lamb_max = lamb_max
        self.lambs = jnp.linspace(self.lamb_min, self.lamb_max, 1000)
        self._precomputed_normalization = None
        
        self.zs = jnp.linspace(0, z_max, 1000)
        self.dVdz_values = Planck15.differential_comoving_volume(self.zs).value * 4 * np.pi

    def dVdz(self, data):
        """Compute the differential comoving volume.

        Args:
            data (dict): Dictionary containing the redshift data.

        Returns:
            numpy.ndarray: Array of differential comoving volume values.
        """
        return jnp.interp(data[self.var_name], self.zs, self.dVdz_values)

    def _normalization(self, params):
        """Compute the normalization factor.

        Args:
            params (dict): Dictionary containing the hyperparameter value.

        Returns:
            float: Normalization factor.
        """
        psi_of_z = self.psi_of_z({self.var_name: self.zs}, params)
        norm = jax.scipy.integrate.trapezoid(psi_of_z * self.dVdz_values / (1 + self.zs), self.zs)
        return norm

    def precomputed_normalization_array(self):
        """Precompute the normalization array."""
        if self._precomputed_normalization is None:
            self._precomputed_normalization = jnp.array([self._normalization({'lamb':lamb}) for lamb in self.lambs])
            return self._precomputed_normalization
        else:
            return self._precomputed_normalization

    def normalization(self, params):
        """Compute the normalization factor for given parameters.

        Args:
            params (dict): Dictionary containing the hyperparameter value.

        Returns:
            float: Normalization factor.
        """
        norms = self.precomputed_normalization_array()
        return jnp.interp(params[self.hyper_var_name], self.lambs, norms)

    def probability(self, data, params):
        """Compute the probability distribution.

        Args:
            data (dict): Dictionary containing the redshift data.
            params (dict): Dictionary containing the hyperparameter value.

        Returns:
            numpy.ndarray: Array of probability values.
        """
        normalisation = self._normalization(params)
        prob = self.dVdz(data) * self.psi_of_z(data, params) / (1 + data[self.var_name])
        in_bounds = box(data[self.var_name], 0, self.z_max)
        #return (prob  / normalisation) * in_bounds
        #print(prob, in_bounds,normalisation)
        #print(prob *in_bounds / normalisation)
        return prob *in_bounds / normalisation
        #return prob * in_bounds# * jnp.abs(params[self.hyper_var_name])

    def __call__(self, data, params):
        """Call the probability method.

        Args:
            data (dict): Dictionary containing the redshift data.
            params (dict): Dictionary containing the hyperparameter value.

        Returns:
            numpy.ndarray: Array of probability values.
        """
        return jnp.nan_to_num(self.probability(data, params),nan=0)

    def total_four_volume(self, params, analysis_time):
        return self.normalization(params) / 1e9 * analysis_time


class PowerLawRedshift(Redshift):
    """A subclass of Redshift representing power law redshift distributions.

    Methods:
        psi_of_z(data, params): Computes the power law function.
    """
    def psi_of_z(self, data, params):
        """Compute the power law function.

        Args:
            data (dict): Dictionary containing the redshift data.
            params (dict): Dictionary containing the hyperparameter value.

        Returns:
            numpy.ndarray: Array of power law values.
        """
        return ((1+data[self.var_name])**(params[self.hyper_var_name]))





import jax
import jax.numpy as jnp
from ..generic import *
from .analytic_models import TruncatedGaussian1DIndependentAnalytic
import pandas as pd
import numpy as np

class MixtureTruncatedGaussian1DFixedZero(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1'], hyper_var_names=['mu_chi_a', 'sigma_chi_a', 'sigma_chi_b', 'eta']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.eta_variable = self.hyper_var_names[-1]
        kwargs = {'a' : a, 'b' : b}
        self.models = [TruncatedGaussian1DAnalytic(var_names=[var_names[0]], 
                                                   hyper_var_names=[hyper_var_names[0], hyper_var_names[1]], 
                                                   **kwargs),
                       TruncatedGaussian1DAnalytic(var_names=[var_names[0]], 
                                                   hyper_var_names=['mu_zero_spin_1d_fixed', hyper_var_names[2]], 
                                                   **kwargs),]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.eta_variable]  * self.models[0](data, params)
        result += (1 - params[self.eta_variable]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.eta_variable]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.eta_variable]) * self.models[1].evaluate(data, params)
        return result



class TruncatedGaussian1DMixtureChi1StandardChi2(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1_a', 'sigma_chi_1_a', 'sigma_chi_1_b', 'eta_1', 'mu_chi_2', 'sigma_chi_2']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        kwargs = {'a' : a, 'b' : b}
        chi_1_var_names = [var_names[0]]
        chi_2_var_names = [var_names[1]]

        chi_1_hyper_var_names = self.hyper_var_names[0:4]
        chi_2_hyper_var_names = self.hyper_var_names[4:6]

        self.models = [MixtureTruncatedGaussian1DFixedZero(var_names=chi_1_var_names, hyper_var_names=chi_1_hyper_var_names, a=a[0], b=b[0]),
                       TruncatedGaussian1DAnalytic(var_names=chi_2_var_names, hyper_var_names=chi_2_hyper_var_names, a=a[1], b=b[1])]

    @property
    def limits(self):
        return {var : [self.a[i], self.b[i]] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result = self.models[0](data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i](data, params)
        return result

    def evaluate(self, data, params):
        result = self.models[0].evaluate(data, params)
        for i in range(1,len(self.models)):
            result *= self.models[i].evaluate(data, params)
        return result



class TruncatedGaussian1DMixtureZeroAndFloating(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['sigma_chi_at_0', 'mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = ['mu_zero_spin_1d_fixed', self.hyper_var_names[0], 'mu_zero_spin_1d_fixed', self.hyper_var_names[0]]
        comp2_hyper_var_names = self.hyper_var_names[1:5]
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=a, b=b)]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        df_hyper_samples['mu_zero_spin_1d_fixed'] = 0
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)


class TruncatedGaussian1DMixtureZeroAndFloatingSpike(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = ['mu_zero_spin_1d_fixed', 'sigma_chi_at_0', 'mu_zero_spin_1d_fixed', 'sigma_chi_at_0']
        comp2_hyper_var_names = self.hyper_var_names[0:4]
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=a, b=b)]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        params['sigma_chi_at_0'] = 1e-6
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        params['sigma_chi_at_0'] = 1e-6
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        df_hyper_samples['mu_zero_spin_1d_fixed'] = 0
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)

class TruncatedGaussian1DMixtureZeroAndUniform(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['sigma_chi_at_0','eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = ['mu_zero_spin_1d_fixed', self.hyper_var_names[0], 'mu_zero_spin_1d_fixed', self.hyper_var_names[0]]
        comp2_hyper_var_names = []
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       Uniform2DAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=[a,a], b=[b,b])]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        df_hyper_samples['mu_zero_spin_1d_fixed'] = 0
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)

class TruncatedGaussian1DMixtureZeroFloatingAndUniform(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2','eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = [self.hyper_var_names[0], self.hyper_var_names[1], self.hyper_var_names[2], self.hyper_var_names[3]]
        comp2_hyper_var_names = []
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       Uniform2DAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=[a,a], b=[b,b])]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)

class TruncatedGaussian2DMixtureZeroFloatingAndUniform(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2','rho_chi', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = [self.hyper_var_names[0], self.hyper_var_names[1], self.hyper_var_names[2], self.hyper_var_names[3], self.hyper_var_names[4]]
        comp2_hyper_var_names = []
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian2DAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       Uniform2DAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=a, b=b)]

    @property
    def limits(self):
        return {var : [self.a[i], self.b[i]] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)


class TruncatedGaussian2DWithLabelSwitching(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2','rho_chi', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = [self.hyper_var_names[0], self.hyper_var_names[1], self.hyper_var_names[2], self.hyper_var_names[3], self.hyper_var_names[4]]
        comp2_hyper_var_names = []
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian2DAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       TruncatedGaussian2DAnalytic(var_names=self.var_names[::-1], hyper_var_names=comp1_hyper_var_names, a=a[::-1], b=b[::-1])]

    @property
    def limits(self):
        return {var : [self.a[i], self.b[i]] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)


class TruncatedGaussian1DWithLabelSwitching(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = [self.hyper_var_names[0], self.hyper_var_names[1], self.hyper_var_names[2], self.hyper_var_names[3]]
        comp2_hyper_var_names = []
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names[::-1], hyper_var_names=comp1_hyper_var_names, a=a, b=b)]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)


class TruncatedGaussian1DMixtureZeroAndFloatingIID(AnalyticPopulationModel, SpinPopulationModel):
    def __init__(self, a, b, var_names=['chi_1', 'chi_2'], hyper_var_names=['sigma_chi_at_0','mu_chi', 'sigma_chi', 'eta_spin']):
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.a = a; self.b=b;
        kwargs = {'a' : a, 'b' : b}
        comp1_var_names = [var_names[0]]
        comp2_var_names = [var_names[1]]

        comp1_hyper_var_names = ['mu_zero_spin_1d_fixed', self.hyper_var_names[0], 'mu_zero_spin_1d_fixed', self.hyper_var_names[0]]
        comp2_hyper_var_names = self.hyper_var_names[1:3]*2
        self.mixture_hyper_var_name = self.hyper_var_names[-1]

        self.models = [TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp1_hyper_var_names, a=a, b=b),
                       TruncatedGaussian1DIndependentAnalytic(var_names=self.var_names, hyper_var_names=comp2_hyper_var_names, a=a, b=b)]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}

    def __call__(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.mixture_hyper_var_name]  * self.models[0](data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1](data, params)
        return result

    def evaluate(self, data, params):
        params['mu_zero_spin_1d_fixed'] = 0.0
        result  =      params[self.mixture_hyper_var_name]  * self.models[0].evaluate(data, params)
        result += (1 - params[self.mixture_hyper_var_name]) * self.models[1].evaluate(data, params)
        return result

    def sample(self, df_hyper_samples_in, oversample=1, **kwargs):
        kwargs['oversample'] = oversample
        df_hyper_samples = df_hyper_samples_in.copy()
        df_hyper_samples['mu_zero_spin_1d_fixed'] = 0
        series_1 = self.models[0].sample(df_hyper_samples, **kwargs)
        series_2 = self.models[1].sample(df_hyper_samples, **kwargs)
        N = len(df_hyper_samples)
        sampled_mixture = np.hstack([(df_hyper_samples[self.mixture_hyper_var_name] <= np.random.rand(N)) for _ in range(oversample)])
        d = {col : ((sampled_mixture * series_1[col]) + ((1 - sampled_mixture) * series_2[col])) for col in series_1.columns}
        return pd.DataFrame(d)




import jax
import jax.numpy as jnp
import os

def is_in_jupyter():
    return "JUPYTER_RUNTIME_DIR" in os.environ


def chunked_vmap(func, in_axes, chunk=100, progress_note=None):
    # Pre-compile the vmap function
    vmap_func = jax.vmap(lambda p: func(p), in_axes)

    from tqdm.auto import tqdm

    # Define a function that applies vmap to input chunks
    def chunk_vmap(params):
        # Get the keys of the input dictionary
        keys = list(params.keys())
        # Get the size of the first key (assuming all keys have the same size)
        size = len(params[keys[0]])

        # Initialize a list to store the results of vmap for each chunk
        results = []

        progress_title = f"{progress_note}" if progress_note else None

        pbar = tqdm(total=size, desc=progress_title)

        # Iterate over the chunks
        for start_idx in range(0, size, chunk):
            end_idx = min(start_idx + chunk, size)

            # Slice the input dictionary to get the chunk
            chunk_params = {key: value[start_idx:end_idx] for key, value in params.items()}

            # Apply vmap to the function with the chunked input
            result = vmap_func(chunk_params)
            results.append(result)

            pbar.update(end_idx - start_idx)

        pbar.close()

        # Concatenate the results along the first axis
        if len(results) == 1:
            return results[0]
        else:
            return jnp.concatenate(results, axis=0)

    # Return the function that applies vmap to input chunks
    return chunk_vmap



from dataclasses import dataclass
from typing import List, Union, Dict, Optional
import jax.numpy as jnp
import jax
import numpy


@dataclass
class Grid1D:
	name : str = 'x'
	minimum : float = 0.0
	maximum : float = 0.0
	N : int = 100
	latex_name : Optional[str] = None

	def __post_init__(self):
		if self.latex_name is None:
			new_name = self.name.replace('_','') 
			self.latex_name = fr'${new_name}$'

		self.grid = jnp.linspace(self.minimum, self.maximum, self.N)
		self._data = {self.name : self.grid}

	@property
	def data(self):
		return self._data
	

@dataclass
class Grid:
	grid_list : List[Grid1D]

	def __post_init__(self):
		self.grid_1ds = {g.name : g.grid for g in self.grid_list}
		grid_combos = jnp.meshgrid(*[g.grid for g in self.grid_list])
		data = {}
		for i,g in enumerate(self.grid_list):
			data[g.name] = grid_combos[i]
		self._data = data

	@property
	def data(self):
		return self._data



import numpy as np
from jax import random
import jax.numpy as jnp
from numpyro.distributions import Distribution, constraints
from numpyro.distributions.util import promote_shapes

class LKJCorrelationPrior(Distribution):
    support = constraints.interval(-1, 1)  # Define the support of the distribution
    has_enumerate_support = False
    low = -1
    high = 1

    def __init__(self):
        super().__init__(batch_shape=(), event_shape=())
        
    @staticmethod
    def real_value(u):
        # Define the complex number
        z = (1 - 1j * np.sqrt(3)) / (2 * (-1 + 2*u + 2j * np.sqrt(u - u**2))**(1/3)) + \
            0.5 * (1 + 1j * np.sqrt(3)) * (-1 + 2*u + 2j * np.sqrt(u - u**2))**(1/3)

        # Compute the real part of the complex number
        real_part = np.real(z)
        return real_part

    def sample(self, key, sample_shape=()):
        # Inverse transform sampling method
        u = random.uniform(key, shape=sample_shape)
        return self.real_value(u)  # Sampling from the inverse CDF of the distribution

    def log_prob(self, value):
        # Ensure the value is within the support
        value = promote_shapes(value)[0]
        valid = (value >= -1) & (value <= 1)
        log_prob = jnp.where(
            valid,
            jnp.log(3 / 4) + jnp.log(1 - value ** 2),
            -jnp.inf
        )
        return log_prob

    @property
    def mean(self):
        return 0.0

    @property
    def variance(self):
        return 4 / 45  # Variance of the distribution


import jax
import jax.numpy as jnp
from jax import random
from numpyro.distributions import Distribution, constraints

class DiracDelta(Distribution):
    support = constraints.real
    has_enumerate_support = False

    def __init__(self, value):
        self.value = value
        batch_shape = ()
        event_shape = ()
        super().__init__(batch_shape=batch_shape, event_shape=event_shape)

    def sample(self, key, sample_shape=()):
        return jnp.full(sample_shape, self.value)

    def log_prob(self, value):
        log_prob = jnp.where(value == self.value, 0.0, -jnp.inf)
        return log_prob

    @property
    def mean(self):
        return self.value

    @property
    def variance(self):
        return 0.0

import numpy as np
import jax.numpy as jnp
import numpyro
from numpyro.distributions import constraints
from numpyro.handlers import seed, trace
import matplotlib.pyplot as plt

class Triangular(Distribution):
    support = constraints.interval(0, 1)
    
    def __init__(self, a=0.0, b=1.0, mode=0.5, validate_args=None):
        self.a, self.b, self.mode = a, b, mode
        super().__init__(batch_shape=(), event_shape=(), validate_args=validate_args)
    
    def sample(self, key, sample_shape=()):
        u = numpyro.distributions.Uniform(0, 1).sample(key, sample_shape=sample_shape)
        condition = (u < (self.mode - self.a) / (self.b - self.a))
        x = jnp.where(condition,
                      self.a + jnp.sqrt(u * (self.b - self.a) * (self.mode - self.a)),
                      self.b - jnp.sqrt((1 - u) * (self.b - self.a) * (self.b - self.mode)))
        return x

    def log_prob(self, value):
        a, b, mode = self.a, self.b, self.mode
        norm = 2 / (b - a)
        condition1 = (value >= a) & (value < mode)
        condition2 = (value >= mode) & (value <= b)
        result = jnp.where(condition1, norm * (value - a) / (mode - a), 0)
        result = jnp.where(condition2, norm * (b - value) / (b - mode), result)
        return jnp.log(result)








from dataclasses import dataclass, field
from typing import List, Union, Dict, Optional
from ..models.spin import *
from ..models.utils import box
import jax
import jax.numpy as jnp

def alpha_beta_max_to_mu_var_max(alpha, beta, amax):
    mu = alpha / (alpha + beta) * amax
    var = alpha * beta / ((alpha + beta) ** 2 * (alpha + beta + 1)) * amax ** 2
    return mu, var, amax


def mu_var_max_to_alpha_beta_max(mu, var, amax):
    mu /= amax
    var /= amax ** 2
    alpha = (mu ** 2 * (1 - mu) - mu * var) / var
    beta = (mu * (1 - mu) ** 2 - (1 - mu) * var) / var
    return alpha, beta, amax


@dataclass
class AlphaBetaConstraint:
	hyper_var_names : List[str] = field(default_factory=lambda : ['mu_chi', 'sigma_chi'])
	alpha : List[int] = field(default_factory=lambda : [1,1e5])
	beta : List[int] = field(default_factory=lambda : [1,1e5])

	def logpdf(self, x):
		mu = x[self.hyper_var_names[0]]
		var = x[self.hyper_var_names[1]]
		amax = x.get('amax', 1)

		alpha, beta, _ = mu_var_max_to_alpha_beta_max(mu, var, amax)

		is_alpha_within_constraint = (alpha > self.alpha[0]) & (alpha < self.alpha[1])
		is_beta_within_constraint = (beta > self.beta[0]) & (beta < self.beta[1])
		both_within_constraint = is_alpha_within_constraint & is_beta_within_constraint

		return jnp.log(box(alpha, self.alpha[0], self.alpha[1]))  + jnp.log(box(beta, self.beta[0], self.beta[1]))


@dataclass
class FlatInMeanVar:
    hyper_var_names : List[str] = field(default_factory=lambda : ['mu_chi', 'sigma_chi'])
    
    def mean_var_Z_trunc_norm(self, mu, sigma):
        alpha,  beta = (-mu/sigma), ((1-mu)/sigma)
        Z = jax.scipy.stats.norm.cdf(beta) -  jax.scipy.stats.norm.cdf(alpha)
        kappa = (jax.scipy.stats.norm.pdf(beta) - jax.scipy.stats.norm.pdf(alpha)) / Z
        nu = (beta*jax.scipy.stats.norm.pdf(beta) - alpha*jax.scipy.stats.norm.pdf(alpha)) / Z
        E = mu * (1 + kappa/alpha)
        V = (sigma**2) * (1 - nu - kappa**2)
        return E, V, Z

    def log_prior_mean_var_trunc_norm(self, mu, sigma):
        E, V, Z = self.mean_var_Z_trunc_norm(mu, sigma)
        log_prob = 5*jnp.log(Z)
        log_prob += -0.5*((jnp.log(V)-jnp.log(0.015))/(1.5))**2
        log_prob +=  -0.1*(0.5*((V-0.04)/(0.01))**2 + 0.5*((E-0.5)/(0.3))**2)
        return log_prob

    def logpdf(self, x):
        mu = x[self.hyper_var_names[0]]
        sigma = x[self.hyper_var_names[1]]
        return self.log_prior_mean_var_trunc_norm(mu, sigma)

@dataclass
class FlatInMeanStd:
    hyper_var_names : List[str] = field(default_factory=lambda : ['mu_chi', 'sigma_chi'])
    
    def mean_var_Z_trunc_norm(self, mu, sigma):
        alpha,  beta = (-mu/sigma), ((1-mu)/sigma)
        Z = jax.scipy.stats.norm.cdf(beta) -  jax.scipy.stats.norm.cdf(alpha)
        kappa = (jax.scipy.stats.norm.pdf(beta) - jax.scipy.stats.norm.pdf(alpha)) / Z
        nu = (beta*jax.scipy.stats.norm.pdf(beta) - alpha*jax.scipy.stats.norm.pdf(alpha)) / Z
        E = mu * (1 + kappa/alpha)
        V = (sigma**2) * (1 - nu - kappa**2)
        return E, V, Z

    def log_prior_mean_std_trunc_norm(self, mu, sigma):
        E, V, Z = self.mean_var_Z_trunc_norm(mu, sigma)
        log_prob = 5*jnp.log(Z)
        log_prob += -0.5*((jnp.log(V)-jnp.log(0.015))/(1.5))**2
        log_prob +=  -0.1*(0.5*((V-0.04)/(0.01))**2 + 0.5*((E-0.5)/(0.3))**2)
        return log_prob - 0.5*jnp.log(V)

    def logpdf(self, x):
        mu = x[self.hyper_var_names[0]]
        sigma = x[self.hyper_var_names[1]]
        return self.log_prior_mean_std_trunc_norm(mu, sigma)


@dataclass
class VarianceToStd:
    hyper_var_names : List[str] = field(default_factory=lambda : ['sigma_chi'])
    
    def logpdf(self, x):
        sigma = x[self.hyper_var_names[0]]
        return -jnp.log(sigma)





#########################################################################
### Truncated 2D Gaussian
########################################################################

from ..utils import *
from .abstract import *
from ..sample import *

import jax
import jax.numpy as jnp
from typing import Union, List, Dict
from dataclasses import dataclass, field

@dataclass(frozen=True)
class CovarianceMatrix2D:
    s1  : Union[jax.Array,float]
    s2  : Union[jax.Array,float]
    rho : Union[jax.Array,float]
    
    def inv(self):
        rhom1 = jnp.sqrt(1-self.rho**2)
        s1_ = 1 / ( self.s1 * rhom1 )
        s2_ = 1 / ( self.s2 * rhom1 )
        rho_ = -self.rho
        return CovarianceMatrix2D(s1_, s2_, rho_)
    
    def __add__(self, other):
        s1_ = jnp.sqrt(self.s1**2 + other.s1**2)
        s2_ = jnp.sqrt(self.s2**2 + other.s2**2)
        rho_ =  self.rho * self.s1 * self.s2
        rho_ += other.rho * other.s1 * other.s2
        rho_ /= s1_ * s2_
        return CovarianceMatrix2D(s1_, s2_, rho_)
    
    def unpack(self):
        return self.s1, self.s2, self.rho
    
    @property
    def value(self):
        s1,s2,rho = self.unpack()
        return jnp.array([[s1**2, s1*s2*rho],[s1*s2*rho, s2**2]])
    
    
@dataclass(frozen=True)
class Vector2D:
    x1  : Union[jax.Array,float]
    x2  : Union[jax.Array,float]
    
    def __rmul__(self, cov_matrix):
        s1, s2, rho = cov_matrix.unpack()
        x1_ = s1**2 * self.x1 + rho * s1 * s2 * self.x2
        x2_ = rho * s1 * s2 * self.x1 + s2**2 * self.x2
        return Vector2D(x1_, x2_)
    
    def __add__(self, vector):
        return Vector2D(self.x1 + vector.x1, self.x2 + vector.x2)
    
    @property
    def value(self):
        x1 = self.x1
        x2 = self.x2
        return jnp.array([x1, x2])
    
    
    
def probability_mass(Mu, Sigma, a=[0,0], b=[1,1]):
    return mvnorm2d(Mu.x1, Mu.x2, Sigma.s1, Sigma.s2, a[0], a[1], b[0], b[1], Sigma.rho)

@dataclass(frozen=True)
class TruncatedNormal2D:
    mu : Vector2D
    sigma : CovarianceMatrix2D
    a : List[float] = field(default_factory=lambda: [0.0, 0.0])
    b : List[float] = field(default_factory=lambda: [1.0, 1.0])
    
    def probability_mass(self):
        return probability_mass(self.mu, self.sigma, a=self.a, b=self.b)



def transform_component(mu_1, sigma_1, mu_2, sigma_2):
    Lambda_1 = sigma_1.inv()
    Lambda_2 = sigma_2.inv()
    sigma_ = (Lambda_1 + Lambda_2).inv()    
    mu_ = sigma_ * ( Lambda_1 * mu_1 + Lambda_2 * mu_2 )
    return mu_, sigma_

@jit                  
def pack_values(mu_11=0.0, mu_12=0.0, mu_21=0.0, mu_22=0.0,
                       sigma_11=1.0, sigma_12=1.0, sigma_21=1.0, sigma_22=1.0,
                       rho_1=0.0, rho_2=0.0):
    mu_1 = Vector2D(mu_11, mu_12)
    sigma_1 = CovarianceMatrix2D(sigma_11, sigma_12, rho_1)
    mu_2 = Vector2D(mu_21, mu_22)
    sigma_2 = CovarianceMatrix2D(sigma_21, sigma_22, rho_2)
    return mu_1, sigma_1, mu_2, sigma_2

def make_cov(sig_1, sig_2, rho):
    return jnp.stack([jnp.stack([sig_1**2, sig_1*sig_2*rho],axis=-1)
                     ,jnp.stack([sig_1*sig_2*rho, sig_2**2],axis=-1)], axis=-1)


multivariate_normal_pdf = jax.vmap(jax.scipy.stats.multivariate_normal.pdf, in_axes=(0, 0, 0))

def compute_normal_pdf(mu_11, mu_12, sig_11, sig_12, rho_1,
                        mu_21, mu_22, sig_21, sig_22, rho_2):
    #argss = (mu_11, mu_12, sig_11, sig_12, rho_1, mu_21, mu_22, sig_21, sig_22, rho_2)
    #for a in argss:
    #    print(a)
    mu_1 = jnp.stack([mu_11, mu_12], axis=-1);
    mu_2 = jnp.stack([mu_21, mu_22], axis=-1);
    sig_1 = make_cov(sig_11, sig_12, rho_1)#jnp.stack([jnp.stack([sig_11**2, sig_11*sig_12*rho_1]),jnp.stack([sig_11*sig_12*rho_1, sig_12**2])])
    sig_2 = make_cov(sig_21, sig_22, rho_2)#jnp.stack([jnp.stack([sig_21**2, sig_21*sig_22*rho_2]),jnp.stack([sig_21*sig_22*rho_2, sig_22**2])])
    return jax.scipy.stats.multivariate_normal.pdf(mu_1, mu_2, sig_1 + sig_2)


@jit
def compute_likelihood(mu_11, mu_12, mu_21, mu_22,
                       sigma_11, sigma_12, sigma_21, sigma_22,
                       rho_1, rho_2, a=[0.0, 0.0], b=[1.0, 1.0]):
    mu_1 = Vector2D(mu_11, mu_12)
    sigma_1 = CovarianceMatrix2D(sigma_11, sigma_12, rho_1)
    mu_2 = Vector2D(mu_21, mu_22)
    sigma_2 = CovarianceMatrix2D(sigma_21, sigma_22, rho_2)
    mu_, sigma_ = transform_component(mu_1, sigma_1, mu_2, sigma_2)
    final_prob  = probability_mass(mu_, sigma_, a=a, b=b)
    final_prob /= probability_mass(mu_1, sigma_1, a=a, b=b)
    final_prob /= probability_mass(mu_2, sigma_2, a=a, b=b)
    ## add part to compute loglikelihood here 
    final_prob *= compute_normal_pdf(mu_1.x1, mu_1.x2, sigma_1.s1, sigma_1.s2, sigma_1.rho,
                                     mu_2.x1, mu_2.x2, sigma_2.s1, sigma_2.s2, sigma_2.rho)
    return final_prob

@jit
def compute_likelihood_fixed_data_limits(mu_11, mu_12, mu_21, mu_22,
                       sigma_11, sigma_12, sigma_21, sigma_22,
                       rho_1, rho_2, a=[0.0, 0.0], b=[1.0, 1.0], a_data=[0.0, 0.0], b_data=[1.0, 1.0]):
    mu_1 = Vector2D(mu_11, mu_12)
    sigma_1 = CovarianceMatrix2D(sigma_11, sigma_12, rho_1)
    mu_2 = Vector2D(mu_21, mu_22)
    sigma_2 = CovarianceMatrix2D(sigma_21, sigma_22, rho_2)
    mu_, sigma_ = transform_component(mu_1, sigma_1, mu_2, sigma_2)
    final_prob  = probability_mass(mu_, sigma_, a=a, b=b)
    final_prob /= probability_mass(mu_1, sigma_1, a=a_data, b=b_data)
    final_prob /= probability_mass(mu_2, sigma_2, a=a, b=b)
    ## add part to compute loglikelihood here 
    final_prob *= compute_normal_pdf(mu_1.x1, mu_1.x2, sigma_1.s1, sigma_1.s2, sigma_1.rho,
                                     mu_2.x1, mu_2.x2, sigma_2.s1, sigma_2.s2, sigma_2.rho)
    return final_prob



@jit
def compute_normal_pdf_on_samples(x_1, x_2, mu_1, mu_2, sigma_1, sigma_2, rho):
    Z = jax.scipy.stats.multivariate_normal.pdf( 
                    jnp.stack([x_1, x_2], axis=-1),
                    jnp.stack([mu_1, mu_2], axis=-1),
                    jnp.stack([jnp.stack([sigma_1**2, sigma_1*sigma_2*rho]),jnp.stack([sigma_1*sigma_2*rho, sigma_2**2])], axis=-1))
    return Z


@jit
def compute_truncated_normal_pdf_on_samples(x_1, x_2, mu_1, mu_2, sigma_1, sigma_2, rho, a=[0.0, 0.0], b=[1.0, 1.0]):
    Z = jax.scipy.stats.multivariate_normal.pdf( 
                    jnp.stack([x_1, x_2], axis=-1),
                    jnp.stack([mu_1, mu_2], axis=-1),
                    jnp.stack([jnp.stack([sigma_1**2, sigma_1*sigma_2*rho]),jnp.stack([sigma_1*sigma_2*rho, sigma_2**2])], axis=-1))
    
    Z = Z / mvnorm2d(mu_1, mu_2, sigma_1, sigma_2, a[0], a[1], b[0], b[1], rho)
    return Z
    

class TruncatedGaussian2DAnalytic(AnalyticPopulationModel):
    def __init__(self, a=[0.0, 0.0], b=[1.0, 1.0], var_names=['chi_1', 'chi_2'], hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2', 'rho_chi']):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.var_name_1 = var_names[0]
        self.mu_name_1 = hyper_var_names[0]
        self.sigma_name_1 = hyper_var_names[1]
        self.var_name_2 = var_names[1]
        self.mu_name_2 = hyper_var_names[2]
        self.sigma_name_2 = hyper_var_names[3]
        self.rho_name = hyper_var_names[4]

    @property
    def limits(self):
        return {var : [self.a[i], self.b[i]] for i,var in enumerate(self.var_names)}
        
    def get_data(self, data, params, component):
        var_name = self.var_name_1 if (component == 1) else self.var_name_2
        mu_name = self.mu_name_1 if (component == 1) else self.mu_name_2
        sigma_name = self.sigma_name_1 if (component == 1) else self.sigma_name_2
        X_locations = data[var_name + '_mu_kernel'];
        X_scales    = data[var_name + '_sigma_kernel'];
        mu          = params[mu_name]
        sigma       = params[sigma_name]
        return X_locations, X_scales, mu, sigma
    
    def evaluate(self, data, params):
        x_1, x_2 = data[self.var_name_1], data[self.var_name_2]
        mu_1, mu_2 = params[self.mu_name_1], params[self.mu_name_2]
        sigma_1, sigma_2 = params[self.sigma_name_1], params[self.sigma_name_2]
        rho = params.get(self.rho_name, 1e-6)
        return compute_truncated_normal_pdf_on_samples(x_1, x_2, mu_1, mu_2, sigma_1, sigma_2, rho, a=self.a, b=self.b)
    
    def __call__(self, data, params):
        X_locations_1, X_scales_1, mu_1, sigma_1 = self.get_data(data, params, component=1);
        X_locations_2, X_scales_2, mu_2, sigma_2 = self.get_data(data, params, component=2);
        rho = params.get(self.rho_name, 1e-6)
        rho_kernel = data.get(self.var_name_1 + "_rho_kernel", 1e-6 * jnp.ones_like(data[self.var_name_1 + "_mu_kernel"]))
        return compute_likelihood(mu_11=X_locations_1, mu_12=X_locations_2, 
                                   mu_21=mu_1, mu_22=mu_2,
                                   sigma_11=X_scales_1, sigma_12=X_scales_2, 
                                   sigma_21=sigma_1, sigma_22=sigma_2,
                                   rho_1=rho_kernel, rho_2=rho, a=self.a, b=self.b)

    def sample(self, df_hyper_samples, oversample=1):
        return ppd_truncCorrelatedanalytic(self, df_hyper_samples, oversample=oversample)


class TruncatedGaussian2DAnalyticLimits(AnalyticPopulationModel):
    def __init__(self, a=[0.0, 0.0], b=[1.0, 1.0], var_names=['chi_1', 'chi_2'], 
                 hyper_var_names=['mu_chi_1', 'sigma_chi_1', 'mu_chi_2', 'sigma_chi_2', 'rho_chi',
                                  'chi_1_min', 'chi_1_max', 'chi_2_min', 'chi_2_max']):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.var_name_1 = var_names[0]
        self.mu_name_1 = hyper_var_names[0]
        self.sigma_name_1 = hyper_var_names[1]
        self.var_name_2 = var_names[1]
        self.mu_name_2 = hyper_var_names[2]
        self.sigma_name_2 = hyper_var_names[3]
        self.rho_name = hyper_var_names[4]
        self.a_name_0 = hyper_var_names[5]
        self.b_name_0 = hyper_var_names[6]
        self.a_name_1 = hyper_var_names[7]
        self.b_name_1 = hyper_var_names[8]

    @property
    def limits(self):
        return {var : [self.a[i], self.b[i]] for i,var in enumerate(self.var_names)}
        
    def get_data(self, data, params, component):
        var_name = self.var_name_1 if (component == 1) else self.var_name_2
        mu_name = self.mu_name_1 if (component == 1) else self.mu_name_2
        sigma_name = self.sigma_name_1 if (component == 1) else self.sigma_name_2
        X_locations = data[var_name + '_mu_kernel'];
        X_scales    = data[var_name + '_sigma_kernel'];
        mu          = params[mu_name]
        sigma       = params[sigma_name]
        return X_locations, X_scales, mu, sigma
    
    def evaluate(self, data, params):
        x_1, x_2 = data[self.var_name_1], data[self.var_name_2]
        mu_1, mu_2 = params[self.mu_name_1], params[self.mu_name_2]
        sigma_1, sigma_2 = params[self.sigma_name_1], params[self.sigma_name_2]
        rho = params.get(self.rho_name, 0)
        a = [params.get(self.a_name_0, self.a[0]), params.get(self.a_name_1, self.a[1])];
        b = [params.get(self.b_name_0, self.b[0]), params.get(self.b_name_1, self.b[1])];
        return compute_truncated_normal_pdf_on_samples(x_1, x_2, mu_1, mu_2, sigma_1, sigma_2, rho, a=a, b=b)
    
    def __call__(self, data, params):
        X_locations_1, X_scales_1, mu_1, sigma_1 = self.get_data(data, params, component=1);
        X_locations_2, X_scales_2, mu_2, sigma_2 = self.get_data(data, params, component=2);
        rho = params.get(self.rho_name, 1e-6)
        rho_kernel = data.get(self.var_name_1 + "_rho_kernel", 1e-6 * jnp.ones_like(data[self.var_name_1 + "_mu_kernel"]))
        a = [params.get(self.a_name_0, self.a[0]), params.get(self.a_name_1, self.a[1])];
        b = [params.get(self.b_name_0, self.b[0]), params.get(self.b_name_1, self.b[1])];
        return compute_likelihood_fixed_data_limits(mu_11=X_locations_1, mu_12=X_locations_2, 
                                   mu_21=mu_1, mu_22=mu_2,
                                   sigma_11=X_scales_1, sigma_12=X_scales_2, 
                                   sigma_21=sigma_1, sigma_22=sigma_2,
                                   rho_1=rho_kernel, rho_2=rho, a=a, b=b, a_data=self.a, b_data=self.b)

    def sample(self, df_hyper_samples, oversample=1):
        return ppd_truncCorrelatedanalytic(self, df_hyper_samples, oversample=oversample)








from dataclasses import dataclass, field
from .custompriors import DiracDelta

from typing import List, Dict, Any, Union, Tuple, Optional

import jax
import jax.numpy as jnp

import numpyro
import numpyro.distributions as dist
from numpyro.infer import MCMC, NUTS

from numpyro.infer import (
    MCMC,
    NUTS,
    init_to_feasible,
    init_to_median,
    init_to_sample,
    init_to_uniform,
    init_to_value,
)

import pandas as pd

import numpyro
import numpy as np

@dataclass
class Sampler:
    priors : Dict[str, dist.Distribution]
    latex_symbols : Dict[str, str]
    likelihood : Any = field(repr=False)
    constraints : List = field(default_factory=(lambda : []))
    num_samples : int = 2000
    num_warmup : int = 1000
    seed : Union[None, int] = None
    target_accept_prob : Union[None, float] = 0.7
    summary : bool = True
    return_samples : bool = False
    max_tree_depth : Union[None, int, Tuple] = 6
    just_prior : bool = False
    N_eff : bool = False
    #validation : bool = True
    
    def __post_init__(self):
        self.x = {}
        self._samples = None
        self.samples = None
        if self.latex_symbols is None:
            self.latex_symbols = {k : k for k in self.priors.keys()}
        if self.constraints is None:
            self.constraints = []
    
    def model(self):
        for var,dist in self.priors.items():
            if type(dist) == DiracDelta:
                self.x[var] = dist.value
            else:
                self.x[var] = numpyro.sample(var, dist)

        if len(self.constraints) != 0:
            for i in range(len(self.constraints)):
                numpyro.factor('user_defined_constraint_' + str(i), self.constraints[i].logpdf(self.x))


        if self.N_eff:
            numpyro.deterministic('Selection N_eff', self.likelihood.compute_selection_N_eff_only(self.x))
            numpyro.deterministic('Min Event N_eff', jnp.min(self.likelihood.compute_event_N_eff_only(self.x)))


        if self.just_prior:
            return None

        numpyro.factor('logP', self.likelihood.logpdf(self.x))
        
    def sample(self):
        # Start from this source of randomness. We will split keys for subsequent operations.
        if self.seed is None:
            self.seed = np.random.randint(2*30)
        rng_key = jax.random.PRNGKey(self.seed)
        rng_key, rng_key_ = jax.random.split(rng_key)

        #numpyro.enable_validation(True)

        # Run NUTS.
        kernel = NUTS(self.model, target_accept_prob=self.target_accept_prob, max_tree_depth=self.max_tree_depth)
        num_samples = self.num_samples
        mcmc = MCMC(kernel, num_warmup=self.num_warmup, num_samples=num_samples, jit_model_args=True)
        mcmc.run(rng_key_)
        if self.summary:
            mcmc.print_summary()
        self._samples = mcmc.get_samples()
        
        self.samples = pd.DataFrame(self._samples)
        #if self.latex_symbols:
        #    self.samples.rename(self.latex_symbols, inplace=True, axis=1)
        if self.return_samples:
            return self.samples
        
    def corner(self, color='k', truth=None, truth_color='b'):
        import corner
        figure = corner.corner(self.samples.values,
                                labels=[self.latex_symbols[col] for col in self.samples.columns],
                                quantiles=[0.16, 0.5, 0.84],
                                show_titles=True,
                                color=color,
                                title_kwargs={"fontsize": 12})

        if truth is not None:
            corner.overplot_points(figure, truth[None], marker="s", color=truth_color)
        return figure
        
    def corner_on_fig(self, fig, color='k', truth=None, truth_color='b'):
        import corner
        figure = corner.corner(self.samples.values,
                                labels=[self.latex_symbols[col] for col in self.samples.columns],
                                quantiles=[0.16, 0.5, 0.84],
                                show_titles=True,
                                color=color,
                                title_kwargs={"fontsize": 12}, 
                                fig=fig)
        if truth is not None:
            corner.overplot_points(figure, truth[None], marker="s", color=truth_color)
        return figure


from typing import Union, Dict, List, Optional
from dataclasses import dataclass, field
import pandas as pd
from .grid import Grid1D, Grid
from ..utils.vmap import chunked_vmap
from ..models import AbstractPopulationModel
from ..hyper import PopulationLikelihood
import jax
import jax.numpy as jnp
import numpy as np

DEFAULT_GRID =  Grid([Grid1D(name='chi_1', minimum=0, maximum=1, N=100, latex_name=r"$\chi_1$"), 
                      Grid1D(name='chi_2', minimum=0, maximum=1, N=100, latex_name=r"$\chi_2$")])


CHI_DEFAULT_GRID =  Grid([Grid1D(name='chi_1', minimum=0, maximum=1, N=100, latex_name=r"$\chi_1$"), 
                      Grid1D(name='chi_2', minimum=0, maximum=1, N=100, latex_name=r"$\chi_2$")])

A_DEFAULT_GRID =  Grid([Grid1D(name='a_1', minimum=0, maximum=1, N=100, latex_name=r"$a_1$"), 
                      Grid1D(name='a_2', minimum=0, maximum=1, N=100, latex_name=r"$a_2$")])


latex_labels = {'chi_1' : r'$\chi_1$', 'chi_2' : r'$\chi_2$', 'a_1' : r'$a_1$', 'a_2' : r'$a_2$'}


def has_spin_named_chi(posterior):
    if isinstance(posterior, pd.DataFrame):
        cols = list(posterior.columns)
    else:
        cols = list(posterior.keys())
    return any(('chi' in col) for col in cols)

def get_spin_mag_grid(posterior):
    if has_spin_named_chi(posterior):
        return CHI_DEFAULT_GRID
    else:
        return A_DEFAULT_GRID 


@dataclass
class SpinMagintudePlot:
    hyper_posterior_samples : Dict[str, jax.Array] = field(repr=False)
    model : Optional[AbstractPopulationModel] = None
    spin_grid : Optional[Dict[str, Union[Grid, Grid1D]]] = None
    confidence_interval : float = 0.95
    rate : bool = False
    chunk : int = 20
    n_samples : Optional[int] = 1000
    
    def __post_init__(self):
        acceptable_sample_names = self.model.hyper_var_names + ['rate']
        total_indices = self.hyper_posterior_samples[next(iter(self.hyper_posterior_samples.keys()))].size
        print(total_indices)
        if self.n_samples is not None:
            self.index_sampling = np.random.randint(total_indices, size=self.n_samples)
        else:
            self.index_sampling = np.arange(total_indices)
        self.hyper_posterior_samples = {col:value[self.index_sampling] for col,value in self.hyper_posterior_samples.items() if col in acceptable_sample_names}
        #self.hyper_posterior_samples = {col:value for col,value in self.hyper_posterior_samples.items() if col in acceptable_sample_names}
        self._shapes = {key:0 for key in self.hyper_posterior_samples.keys()}
        self.spin_grid = self.spin_grid or get_spin_mag_grid(self.hyper_posterior_samples)
        data = self.spin_grid.data
        self.result = None
        progress_title = "Computing Spin Magnitude Model on the Grid"
        self._vmapped_func = chunked_vmap(lambda x: self.model.evaluate(data, x), in_axes=(self._shapes,), chunk=self.chunk, progress_note=progress_title)
        self.conf = ((1-self.confidence_interval)/2, self.confidence_interval + (1-self.confidence_interval)/2)

        spin_1_name = [m.name for m in self.spin_grid.grid_list if "_1" in m.name][0]
        spin_2_name = [m.name for m in self.spin_grid.grid_list if "_2" in m.name][0]
        self.spin_1_name = spin_1_name
        self.spin_2_name = spin_2_name

        self.spin_1_grid_obj = [m for m in self.spin_grid.grid_list if "_1" in m.name][0]
        self.spin_2_grid_obj = [m for m in self.spin_grid.grid_list if "_2" in m.name][0]
        
        self.spin_1_grid = self.spin_grid.grid_1ds[spin_1_name]
        self.spin_2_grid = self.spin_grid.grid_1ds[spin_2_name]

    def compute(self):
        self.result = self._vmapped_func(self.hyper_posterior_samples)
        if self.rate:
            self.result = self.result * self.hyper_posterior_samples["rate"][..., None, None]

    def compute_if_not_computed(self):
        if self.result is None:
            self.compute()

    def spin_2D_plot(self, quantiles=[0.9, 0.5], dpi=200):
        from truncnormkde import BoundedKDEPlot
        quantiles.sort(reverse=True)
        hyper_posterior_samples = pd.DataFrame(self.hyper_posterior_samples)
        oversample = 10000 // len(hyper_posterior_samples)
        dfsamps = self.model.sample(hyper_posterior_samples, oversample=oversample)

        return BoundedKDEPlot(dfsamps, latex_labels=latex_labels).plot(quantiles=quantiles, dpi=dpi);


    def spin_2D_plot_old(self, quantiles=[0.05, 0.5, 0.95], dpi=200):
        self.compute_if_not_computed()
        spin_all = self.result
        spin_1s = self.spin_grid.data[self.spin_1_name]
        spin_2s = self.spin_grid.data[self.spin_2_name]
        #spin_median = jnp.quantile(spin_all, 0.5, axis=0)
        #spin_lower = jnp.quantile(spin_all, quantiles[0], axis=0)
        #spin_upper = jnp.quantile(spin_all, quantiles[1], axis=0)

        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(ncols=len(quantiles), figsize=(5*(len(quantiles)),5), dpi=dpi)
        for i in range(len(quantiles)):
            contour_plot = axes[i].contourf(spin_1s, spin_2s, jnp.quantile(spin_all, quantiles[i], axis=0))
            axes[i].set_title(f"{np.round(quantiles[i]*100, 2)} % Posterior Predictive")
            plt.colorbar(contour_plot, ax=axes[i])
            axes[i].set_ylabel(r"$\chi_2$")
            axes[i].set_xlabel(r"$\chi_1$")

        return fig

    def spin_2D_plot_old_mean(self, levels=None, log_scale=False, dpi=200, func_to_use=jnp.mean):
        self.compute_if_not_computed()
        spin_all = self.result
        spin_1s = self.spin_grid.data[self.spin_1_name]
        spin_2s = self.spin_grid.data[self.spin_2_name]
        #spin_median = jnp.quantile(spin_all, 0.5, axis=0)
        #spin_lower = jnp.quantile(spin_all, quantiles[0], axis=0)
        #spin_upper = jnp.quantile(spin_all, quantiles[1], axis=0)

        import matplotlib.pyplot as plt

        fig, axes = plt.subplots(1, figsize=(5,5), dpi=dpi)
        if log_scale:
            contour_plot = axes.contourf(spin_1s, spin_2s, jnp.log(func_to_use(spin_all, axis=0)), levels=levels)
        else:
            contour_plot = axes.contourf(spin_1s, spin_2s, func_to_use(spin_all, axis=0), levels=levels)
        axes.set_title(f"Posterior Predictive")
        plt.colorbar(contour_plot, ax=axes)
        axes.set_ylabel(r"$\chi_2$")
        axes.set_xlabel(r"$\chi_1$")

        return fig


    def spin_marginal_plot(self, spin_1_or_2=1, generic=False, ax=None, aspect=0.5, lower=0, color=None, label=None, alpha=0.3):
        self.compute_if_not_computed()
        
        if spin_1_or_2 == 1:
            spin_a = self.spin_1_grid
            spin_b = self.spin_2_grid
            latex_name = self.spin_1_grid_obj.latex_name
            if generic:
                latex_name = latex_name.replace("_1", "")
            axis_to_integrate = 1
            axis_to_aggregate = 0
        else:
            spin_a = self.spin_2_grid
            spin_b = self.spin_1_grid
            latex_name = self.spin_2_grid_obj.latex_name
            if generic:
                latex_name = latex_name.replace("_2", "")
            axis_to_integrate = 2
            axis_to_aggregate = 0
        spin_all = jax.scipy.integrate.trapezoid(self.result, spin_b, axis=axis_to_integrate)
        spin_median = jnp.quantile(spin_all, 0.5, axis=axis_to_aggregate)
        spin_lower = jnp.quantile(spin_all, self.conf[0], axis=axis_to_aggregate)
        spin_upper = jnp.quantile(spin_all, self.conf[1], axis=axis_to_aggregate)
        
        highest, lowest = spin_upper.max(),  max(spin_lower.min(), lower)
        high_x, low_x = spin_a.max(), spin_a.min()
        
        import matplotlib.pyplot as plt
        if ax is None:
            fig,ax = plt.subplots(1)
        ax.plot(spin_a, spin_median, color=color, label=label)
        ax.fill_between(spin_a, spin_lower, spin_upper, alpha=alpha, color=color)
        #ax.set_yscale("log")
        ax.set_xlabel(latex_name)
        if self.rate:
            if generic:
                ax.set_ylabel(r"$\frac{dN}{d\chi}$")
            else:
                ax.set_ylabel(r"$\frac{dN}{d\chi" + str(spin_1_or_2) + r"}$")
        else:
            if generic:
                ax.set_ylabel(r"$p(\chi | \Lambda)$")
            else:
                ax.set_ylabel(r"$p(\chi_" + str(spin_1_or_2) + r" | \Lambda)$")
        ax.grid(True, which='major', linestyle='dotted', linewidth='0.5', color='gray', alpha=0.5)
        ax.set_xlim(spin_a.min(), spin_a.max())
        #ax.set_ylim(bottom=10**(lowest))
        #ax.set_aspect(aspect*(high_x - low_x)/(highest-lowest))
        new_aspect = aspect*(high_x - low_x)/(highest-lowest)
        if (new_aspect > 0) and not (np.isinf(new_aspect)):
            ax.set_aspect(aspect*(high_x - low_x)/(highest-lowest))
        #ax.set_aspect(aspect)

        return ax

    def spin_1_marginal_plot(self, *args, **kwargs):
        return self.spin_marginal_plot(*args, **kwargs, spin_1_or_2=1, generic=False)

    def spin_2_marginal_plot(self, *args, **kwargs):
        return self.spin_marginal_plot(*args, **kwargs, spin_1_or_2=2, generic=False)

    def spin_generic_marginal_plot(self, *args, **kwargs):
        return self.spin_marginal_plot(*args, **kwargs, spin_1_or_2=1, generic=True)


class AbstractPopulationModel:
	pass


class SampledPopulationModel(AbstractPopulationModel):
	def evaluate(self, data, params):
		return self.__call__(data, params)

class AnalyticPopulationModel(AbstractPopulationModel):
	pass



class SpinPopulationModel(AbstractPopulationModel):
	pass

class SpinMagnitudePopulationModel(SpinPopulationModel):
	pass

class SpinOrientationPopulationModel(SpinPopulationModel):
	pass

class RedshiftPopulationModel(AbstractPopulationModel):
	pass

class MassPopulationModel(AbstractPopulationModel):
	pass


from typing import Union, Dict, List, Optional
from dataclasses import dataclass, field
import pandas as pd
from .grid import Grid1D, Grid
from ..utils.vmap import chunked_vmap
from ..models import AbstractPopulationModel
from ..hyper import PopulationLikelihood
import jax
import jax.numpy as jnp
import numpy as np

DEFAULT_GRID =  Grid1D(name='redshift', minimum=0, maximum=1.9, N=100, latex_name=r"$z$")

@dataclass
class RedshiftPlot:
    hyper_posterior_samples : Dict[str, jax.Array] = field(repr=False)
    model : Optional[AbstractPopulationModel] = None
    redshift_grid : Optional[Dict[str, Union[Grid, Grid1D]]] = None
    confidence_interval : float = 0.95
    rate : float = False
    chunk : int = 100
    n_samples : int = 1000
    
    def __post_init__(self):
        acceptable_sample_names = self.model.hyper_var_names + ['rate']
        total_indices = self.hyper_posterior_samples[next(iter(self.hyper_posterior_samples.keys()))].size
        self.index_sampling = np.random.randint(total_indices, size=self.n_samples)
        self.hyper_posterior_samples = {col:value[self.index_sampling] for col,value in self.hyper_posterior_samples.items() if col in acceptable_sample_names}
        #self.hyper_posterior_samples = {col:value for col,value in self.hyper_posterior_samples.items() if col in acceptable_sample_names}
        self._shapes = {key:0 for key in self.hyper_posterior_samples.keys()}
        self.redshift_grid = self.redshift_grid or DEFAULT_GRID
        data = self.redshift_grid.data
        compute_rate = lambda data,x : (1+data['redshift'])**(x[self.model.hyper_var_names[0]])
        self._vmapped_func = chunked_vmap( lambda x: self.model.evaluate(data, x), in_axes=(self._shapes,), chunk=self.chunk)
        progress_title = "Computing Redshift Model on the Grid"
        self._vmapped_func_rate = chunked_vmap( lambda x: compute_rate(data, x), in_axes=(self._shapes,), chunk=self.chunk, progress_note=progress_title)
        self.result = None
        self.conf = ((1-self.confidence_interval)/2, self.confidence_interval + (1-self.confidence_interval)/2)

    def compute(self):
        self.result = self._vmapped_func(self.hyper_posterior_samples)
        if self.rate:
            self.result = self.result * self.hyper_posterior_samples["rate"][..., None, None]

    def compute_if_not_computed(self):
        if self.result is None:
            self.compute()
            
    def plot_model(self, ax=None, aspect=0.5, log_lower=-15, color=None, label=None, alpha=0.3):
        self.compute_if_not_computed()
        redshift_x = list(self.redshift_grid.data.values())[0]
        redshift_y_median = jnp.quantile(self.result, 0.5, axis=0)
        redshift_y_lower = jnp.quantile(self.result, self.conf[0], axis=0)
        redshift_y_upper = jnp.quantile(self.result, self.conf[1], axis=0)
        
        highest, lowest = jnp.log10(redshift_y_upper.max()),  max(jnp.log10(redshift_y_lower.min()), log_lower)
        high_x, low_x = redshift_x.max(), redshift_x.min()
        
        import matplotlib.pyplot as plt
        if ax is None:
            fig,ax = plt.subplots(1)
        else:
            fig = ax.get_figure()
        ax.grid(True, which='major', linestyle='dotted', linewidth='0.5', color='gray', alpha=0.5)
        ax.plot(redshift_x, redshift_y_median, color=color, label=label)
        ax.fill_between(redshift_x, redshift_y_lower, redshift_y_upper, alpha=alpha, color=color)
        ax.set_yscale("log")
        ax.set_xlabel(self.redshift_grid.latex_name)
        ax.set_ylabel(r"$p(z | \Lambda)$")
        second_lowest = np.partition(redshift_x, 1)[1]  
        ax.set_xlim(second_lowest, redshift_x.max())
        ax.set_ylim(bottom=10**(log_lower))
        #ax.set_aspect(aspect*(high_x - low_x)/(highest-lowest))
        new_aspect = aspect*(high_x - low_x)/(highest-lowest)
        if (new_aspect > 0) and not (np.isinf(new_aspect)):
            ax.set_aspect(aspect*(high_x - low_x)/(highest-lowest))
        
        return fig
    
    def plot_1pz(self, ax=None, aspect=0.5, log_lower=-15, color=None, label=None, alpha=0.3):
        self.compute_if_not_computed()
        self.result_1pz = self._vmapped_func_rate(self.hyper_posterior_samples)
        redshift_x = list(self.redshift_grid.data.values())[0]
        redshift_y_median = jnp.quantile(self.result_1pz, 0.5, axis=0)
        redshift_y_lower = jnp.quantile(self.result_1pz, self.conf[0], axis=0)
        redshift_y_upper = jnp.quantile(self.result_1pz, self.conf[1], axis=0)
        
        highest, lowest = jnp.log10(redshift_y_upper.max()),  max(jnp.log10(redshift_y_lower.min()), log_lower)
        high_x, low_x = redshift_x.max(), redshift_x.min()
        
        import matplotlib.pyplot as plt
        if ax is None:
            fig,ax = plt.subplots(1)
        else:
            fig = ax.get_figure()
        ax.grid(True, which='major', linestyle='dotted', linewidth='0.5', color='gray', alpha=0.5)
        ax.plot(redshift_x, redshift_y_median, color=color, label=label)
        ax.fill_between(redshift_x, redshift_y_lower, redshift_y_upper, alpha=alpha, color=color)
        ax.set_yscale("log")
        ax.set_xlabel(self.redshift_grid.latex_name)
        ax.set_ylabel(r"$p(z | \Lambda)$")
        second_lowest = np.partition(redshift_x, 1)[1]  
        ax.set_xlim(second_lowest, redshift_x.max())
        ax.set_ylim(bottom=10**(log_lower))
        #ax.set_aspect(aspect*(high_x - low_x)/(highest-lowest))
        new_aspect = aspect*(high_x - low_x)/(highest-lowest)
        if (new_aspect > 0) and not (np.isinf(new_aspect)):
            ax.set_aspect(aspect*(high_x - low_x)/(highest-lowest))
        
        return fig


import numpy as np
import pandas as pd

def generate_trunc_norm_slow(mu_1, sigma_1, mu_2, sigma_2, rho=None, a = [0.0, 0.0], b = [1.0, 1.0], oversampling=200):
    N = len(mu_1)
    if rho is None:
        rho = np.zeros(N)
    inbounds = False
    X = np.zeros((N,2))
    for i in range(N):
        while not (inbounds):
            Z = np.random.randn(2)
            x1,x2 = sigma_1[i] * Z[0] , sigma_2[i] * (rho[i] * Z[0] + np.sqrt(1 - rho[i]**2) * Z[1])
            x1,x2 = mu_1[i] + x1, mu_2[i] + x2 

            inbounds = ( (x1 < b[0]) & (x1 > a[0]) ) & ((x2 < b[1]) & (x2 > a[1]))
        
        X[i,0] = x1
        X[i,1] = x2
        inbounds=False

    return X

def ppd_truncCorrelatedanalytic(model, df, oversample=1):
    df_samps = []
    for _ in range(oversample):
        X = generate_trunc_norm_slow(mu_1 = df[model.mu_name_1].values,
                                     sigma_1 = df[model.sigma_name_1].values,
                                     mu_2 = df[model.mu_name_2].values,
                                     sigma_2 = df[model.sigma_name_2].values,
                                     rho = df[model.rho_name].values,
                                     a = np.ones(2)*model.a, b = np.ones(2)*model.b)

        df_samp = pd.DataFrame(X, columns=model.var_names)
        df_samps.append(df_samp)
        
    return pd.concat(df_samps)

def ppd_truncUncorrelatedanalytic(model, df, oversample=1):
    df_samps = []
    for _ in range(oversample):
        X = generate_trunc_norm_slow(mu_1 = df[model.models[0].mu_name].values,
                                     sigma_1 = df[model.models[0].sigma_name].values,
                                     mu_2 = df[model.models[1].mu_name].values,
                                     sigma_2 = df[model.models[1].sigma_name].values, 
                                     a = np.ones(2)*model.a, b = np.ones(2)*model.b)

        df_samp = pd.DataFrame(X, columns=model.var_names)
        df_samps.append(df_samp)
        
    return pd.concat(df_samps)


import jax.numpy as jnp
import numpy as np
from dataclasses import dataclass
import numpyro.distributions as dist
from typing import Dict
import jax
from gravpop import * 
import matplotlib.pyplot as plt

def test_sampler():

    @dataclass
    class TestLikelihood:
        data: Dict[str, jax.Array]
        var_name : str = 'x'
        mu_name : str = 'mu'
        sigma_name : str = 'sigma'
        
        def logpdf(self, x):
            mu = x['mu']
            sigma = x['sigma']
            return jnp.sum( (-0.5 * (self.data[self.var_name] - mu)**2 / sigma**2)  - jnp.log(sigma) )

    like = TestLikelihood({'x' : jnp.array(np.random.randn(1000))})

    Samp = Sampler(
    priors = {'mu' : dist.Uniform(-3,3),  
              'sigma' : dist.Uniform(0,4)}, 
    latex_symbols = {'mu' : r'$\mu$', 
                     'sigma' : r'$\sigma$'} , 
    likelihood=like)

    Samp.sample()

    fig1 = Samp.corner(color='r', truth=jnp.array([0,1]));
    Samp.corner_on_fig(fig1, color='k', truth=jnp.array([0,1]));
    plt.savefig("./test/test_sampler_plot.png")



from dataclasses import dataclass, field
import pandas as pd
from typing import Optional, List, Union, Dict, Any
import jax
import jax.numpy as jnp
from ..plots import *
from ..models import *
from ..hyper import *
from ..utils import *

BETA_CONVERTER = BetaDistributionConverter()

MASS_MODELS = [SmoothedTwoComponentPrimaryMassRatio]


def is_a_spin_magnitude_model(S_mag):
	return ("spin" in S_mag.__class__.__name__.lower()) and ("magnitude" in S_mag.__class__.__name__.lower())

def is_a_spin_orientation_model(S_orient):
	return ("spin" in S_orient.__class__.__name__.lower()) and ("orientation" in S_orient.__class__.__name__.lower())



@dataclass
class HyperPosterior:
	posterior : Optional[Union[pd.DataFrame, Dict[str, jax.Array]]]
	likelihood : Any
	models : Optional[Dict[str, AbstractPopulationModel]] = field(default=None)
	mass_model : Optional[AbstractPopulationModel] = None
	redshift_model : Optional[AbstractPopulationModel] = None
	spin_magnitude_model : Optional[AbstractPopulationModel] = None
	spin_orientation_model : Optional[AbstractPopulationModel] = None
	rate : bool = False
	LVK_posterior_file : Optional[str] = None

	def __post_init__(self):
		#print("started making posterior")
		if self.models is not None:
			self.mass_model = self.models.get('mass', None)
			self.redshift_model = self.models.get('redshift', None)
			self.spin_magnitude_model = self.models.get('spin_magnitude', None)
			self.spin_orientation_model = self.models.get('spin_orientation', None)

		if self.mass_model is None:
			possible_models = [model for model in self.likelihood.models if any([isinstance(model, test_model) for test_model in MASS_MODELS])]
			if len(possible_models) != 0:
				self.mass_model = possible_models[0]

		if self.redshift_model is None:
			possible_models = [model for model in self.likelihood.models if isinstance(model, Redshift)]
			if len(possible_models) != 0:
				self.redshift_model = possible_models[0]

		if self.spin_magnitude_model is None:
			possible_models = [model for model in self.likelihood.models if is_a_spin_magnitude_model(model)]
			if len(possible_models) != 0:
				self.spin_magnitude_model = possible_models[0]

		if self.spin_orientation_model is None:
			possible_models = [model for model in self.likelihood.models if is_a_spin_orientation_model(model)]
			if len(possible_models) != 0:
				self.spin_orientation_model = possible_models[0]

		if isinstance(self.posterior, pd.DataFrame):
			self.posterior_dict = {col : jnp.array(self.posterior[col].values) for col in self.posterior.columns}
		elif isinstance(self.posterior, Dict):
			self.posterior_dict = self.posterior
			self.posterior = pd.DataFrame(self.posterior)
		else:
			raise ValueError("posterior should be either a pandas dataframe or a dictonary of arrays")

		#print("making plots")
		if self.mass_model is not None:
			self.mass_plot = MassPlot(self.posterior_dict, model=self.mass_model)
		else:
			self.mass_plot = None

		if self.redshift_model is not None:
			self.redshift_plot = RedshiftPlot(self.posterior_dict, model=self.redshift_model)
		else:
			self.redshift_plot = None

		if self.spin_magnitude_model is not None:
			self.spin_magnitude_plot = SpinMagintudePlot(self.posterior_dict, model=self.spin_magnitude_model)
		else:
			self.spin_magnitude_plot = None

		if self.spin_orientation_model is not None:
			self.spin_orientation_plot = SpinOrientationPlot(self.posterior_dict, model=self.spin_orientation_model)
		else:
			self.spin_orientation_plot = None
		#print("finished making posterior")

		if self.LVK_posterior_file:
			self.LVK_posterior = pd.read_csv(self.LVK_posterior_file)
		else:
			self.LVK_posterior = None
		#self.posterior_dict = BETA_CONVERTER.convert_parameters(self.posterior_dict, remove=False)
		#self.posterior = BETA_CONVERTER.convert_parameters(self.posterior, remove=False)
		

	@classmethod
	def from_file(cls, posterior_sample_file, event_data_file, selection_file, models, rate=False, LikelihoodClass=PopulationLikelihood, 
					   SelectionClass=SelectionFunction, samples_per_event=1000, LVK_posterior_file=None, **kwargs):
		posterior_file_extension = posterior_sample_file.split(".")[-1]
		#print("reading...")
		if posterior_file_extension == "csv":
			## Assume it can be read by pandas
			posterior = pd.read_csv(posterior_sample_file)

		#print("creating likelihood...")
		likelihood = LikelihoodClass.from_file(
		            event_data_filename = event_data_file,
		            selection_data_filename = selection_file,
		            models = models,
		            samples_per_event=samples_per_event,
		            **kwargs
		            )

		return cls(posterior, likelihood, rate=rate, LVK_posterior_file=LVK_posterior_file)

	def calculate_selection_N_eff(self, chunk=100):
		N_eff = self.likelihood.selection_data.calculate_N_eff_chunked(self.likelihood, self.posterior_dict, chunk=chunk)
		self.posterior_dict['pdet_n_effective'] = N_eff
		self.posterior.loc[:, 'pdet_n_effective'] = N_eff

	def calculate_event_N_eff_chunked(self, params, chunk=100):
		N_eff_func = lambda params : self.likelihood.compute_event_N_eff_only(params)
		axes_dict_shape = {key:0 for key in params.keys()}
		N_eff_chunked_func = chunked_vmap(N_eff_func, in_axes=(axes_dict_shape,), chunk=chunk, progress_note="Calculating Event N_effective ...")

		N_effs = N_eff_chunked_func(params)

		return N_effs

	def calculate_N_eff_event(self, chunk=100):
		N_eff = self.calculate_event_N_eff_chunked(self.posterior_dict, chunk=chunk)
		self.posterior_dict['min_event_n_effective'] = N_eff
		self.posterior.loc[:, 'min_event_n_effective'] = N_eff

	def calculate_N_eff_selection(self, chunk=100):
		N_eff = self.likelihood.selection_data.calculate_N_eff_chunked(self.likelihood, self.posterior_dict, chunk=chunk)
		self.posterior_dict['pdet_n_effective'] = N_eff
		self.posterior.loc[:, 'pdet_n_effective'] = N_eff

	def calculate_rates(self, chunk=100):
		rate = self.likelihood.selection_data.calculate_rate_for_hyperparameters_chunked(self.likelihood, self.posterior_dict, chunk=chunk)
		self.posterior_dict['rate'] = rate
		self.posterior.loc[:, 'rate'] = rate

	def N_effective_cuts(self, chunk=100, selection=True, events=True):
		any_column = next(iter(self.posterior_dict.keys()))
		safe = jnp.ones(shape=self.posterior_dict[any_column].shape, dtype='bool')
		if events:
			if 'min_event_n_effective' not in self.posterior.columns:
				self.calculate_N_eff_event(chunk=chunk)
			safe = safe | (self.posterior_dict['min_event_n_effective'] > self.likelihood.N_events)

		if selection:
			if 'pdet_n_effective' not in self.posterior.columns:
				self.calculate_N_eff_selection(chunk=chunk)
			safe = safe | (self.posterior_dict['pdet_n_effective'] > 4*self.likelihood.N_events)

		self.posterior_dict_with_cuts = {key:value[safe] for key,value in self.posterior_dict.items()}
		self.posterior_with_cuts = pd.DataFrame(self.posterior_dict_with_cuts)
		return self.posterior_with_cuts

	def N_effective_cuts_events(self, chunk=100):
		if 'min_event_n_effective' not in self.posterior.columns:
			self.calculate_event_N_eff(chunk=chunk)
		safe = (self.posterior_dict['min_event_n_effective'] > self.likelihood.N_events)
		self.posterior_dict_with_cuts = {key:value[safe] for key,value in self.posterior_dict.items()}
		self.posterior_with_cuts = pd.DataFrame(self.posterior_dict_with_cuts)
		return self.posterior_with_cuts




import jax
import jax.numpy as jnp
from ..models.utils import *
from .selection import SelectionFunction
from ..models.redshift import Redshift
from ..utils import *
from .expectation import *
import numpy as np


from dataclasses import dataclass, field
from typing import List, Dict, Optional, Union
from functools import partial


def fix_kernels(mus, sigmas, a, b, lower_lim=-4, upper_lim=4, n=10):
    alphas, betas = (mus - a)/sigmas, (mus - b)/sigmas
    too_far_left = (alphas < lower_lim) & (betas < 0)
    too_far_right = (betas > upper_lim) & (alphas > 0)
    sigmas_fixed = jnp.where(too_far_left | too_far_right, sigmas/jnp.sqrt(n), sigmas) 
    mus_fixed = jnp.where(too_far_left, a - (a-mus)/n, jnp.where(too_far_right, b + (mus-b)/n, mus))
    return mus_fixed, sigmas_fixed

@dataclass
class HybridPopulationLikelihood:
    sampled_models:   List  # List of population models evaluated using monte-carlo
    analytic_models:  List  # List of population models evaluated analytically
    event_data:       Dict[str, jax.Array] = field(repr=False)    # Dictionary of data e.g. {'mass_1' : jnp.array([21.2, 23.0, ...], ....)}
    selection_data:   Optional[Union[Dict[str, jax.Array], SelectionFunction]] = field(default=None, repr=False)   # Dictionary of data e.g. {'mass_1' : jnp.array([21.2, 23.0, ...], ....)}
    event_expectation: Optional[AbstractExpectation] = field(default=HybridEventExpectation())
    selection_expectation: Optional[AbstractExpectation] = field(default=HybridSelectionExpectation())
    enforce_convergence : Optional[bool] = False
    event_names : Optional[List[str]] = None
    selection_sampled_models: Optional[str] = None
    selection_analytic_models: Optional[str] = None
    kernel_fix_iterations : int = 10
    
    def __post_init__(self):
        if not ('weights' in self.event_data):
            raise ValueError("Expected 'weights' key in the event_data dictionary")
        self.N_events = self.event_data['weights'].shape[0]

        if self.selection_sampled_models is None:
            self.selection_sampled_models = self.sampled_models

        if self.selection_analytic_models is None:
            self.selection_analytic_models = self.analytic_models

        if "prior" not in self.event_data:
            keyvalues = list(self.event_data.items())
            largest_shape = keyvalues[np.argmax([len(value.shape) for key,value in keyvalues])][1].shape # most probably a sampled variable
            #largest_shape = self.event_data[key_with_largest_shape].shape
            self.event_data["prior"] = jnp.ones(shape=largest_shape)

        if isinstance(self.selection_data, Dict):
            redshift_model = [model for model in self.sampled_models if isinstance(model, Redshift)]
            redshift_model += [model for model in self.analytic_models if isinstance(model, Redshift)]
            if len(redshift_model) == 0:
                redshift_model = None
            else:
                redshift_model = redshift_model[0]
            self.selection_data = SelectionFunction(self.selection_data, self.analysis_time, self.total_generated, self.redshift_model)
        elif isinstance(self.selection_data, SelectionFunction):
            self.analysis_time = self.selection_data.analysis_time
            self.total_generated = self.selection_data.total_generated
            self.total_detected = self.selection_data.total_detected
            self.detection_ratio = self.selection_data.detection_ratio
        else:
            print("No selection function provided")
            self.selection_data = None

        self._models = self.sampled_models + self.analytic_models

        self.analytic_limits = {}
        for model in self.analytic_models:
            self.analytic_limits.update(model.limits)

        if self.event_names is None:
            self.event_names = []

        for var in self.analytic_limits.keys():
            if self.selection_data is not None:
                sel = self.selection_data.selection_data
                mus, sigmas = fix_kernels(sel[var + '_mu_kernel'], sel[var  + '_sigma_kernel'], self.analytic_limits[var][0], self.analytic_limits[var][1], n=self.kernel_fix_iterations)
                self.selection_data.selection_data[var + '_mu_kernel'], self.selection_data.selection_data[var + '_sigma_kernel'] = mus, sigmas

            eventd = self.event_data
            mus, sigmas = fix_kernels(eventd[var + '_mu_kernel'], eventd[var  + '_sigma_kernel'], self.analytic_limits[var][0], self.analytic_limits[var][1], n=self.kernel_fix_iterations)
            self.event_data[var + '_mu_kernel'], self.event_data[var + '_sigma_kernel'] = mus, sigmas

        #print("Initialized Likelihood with variables:")
        #print(self.event_data.keys())
        #print("With events:")
        #print(self.event_names)

    @property
    def models(self):
        return self._models
    
    @staticmethod
    def log(x):
        return jnp.log(x + 1e-30)

    def total_event_bayes_factors(self, data, params, N=None, detection_ratio=1):
        return self.analytic_event_bayes_factors(self.event_data, params, detection_ratio) \
                + self.sampled_event_bayes_factors(self.event_data, params, N)

    def sampled_compute_log_weights(self, data, params):
        return sum(self.log(model(data, params)) for model in self.sampled_models) - self.log(data["prior"])

    def _compute_sampled_selection_N_eff(self, logweights, weights, N=None):
        log_mu = jax.scipy.special.logsumexp(logweights, axis=-1) - jnp.log(N) 
        mu = jnp.exp(log_mu)  # E x K
        log_var_1 = jax.scipy.special.logsumexp(2*logweights , axis=-1) - 2*jnp.log(N) # E x K
        log_var_2 = 2*log_mu - jnp.log(N)  # E x K
        var = jnp.exp(log_var_1) - jnp.exp(log_var_2) # E x K
        var = jnp.sum(var * weights ,axis=-1) # E
        mu  = jnp.sum(mu  * weights ,axis=-1) # E
        N_eff = mu**2 / var # E
        return N_eff

    def _compute_sampled_event_N_eff(self, logweights, weights, N=None):
        N = N or logweights.shape[-1]
        log_mu = jax.scipy.special.logsumexp(logweights, axis=-1) # E x K
        mu = jnp.exp(log_mu) # E x K
        mu_squared = jnp.exp(jax.scipy.special.logsumexp(2*logweights , axis=-1)) # E x K

        mu = jnp.sum( mu * weights ,axis=-1)
        mu_squared = jnp.sum( mu_squared * weights, axis=-1)
        N_eff = mu**2 / mu_squared # E x K
        N_eff = jnp.min(jnp.exp(log_N_eff), axis=-1)
        return log_mu- jnp.log(N), N_eff
    
    @staticmethod
    def aggregate_kernels(data, loglikes):
        weights = data['weights'] # E x K
        return jax.scipy.special.logsumexp( loglikes + jnp.log(weights), axis=-1) # E

    def analytic_event_bayes_factors(self, data, params, detection_ratio=1):
        if len(self.analytic_models) == 0:
            return 0.0
        loglikes = sum(self.log(model(data, params)) for model in self.analytic_models) # E x K
        return self.aggregate_kernels(data, loglikes) + jnp.log(detection_ratio)
    
    def sampled_event_bayes_factors(self, data, params, N=None):
        if len(self.sampled_models) == 0:
            return 0.0
        loglikes = sum(self.log(model(data, params)) for model in self.sampled_models) # E x K x N
        N = N or loglikes.shape[-1]
        if "prior" not in data.keys():
            log_priors = 0
        else:
            log_priors = self.log(data["prior"])
        loglikes = jax.scipy.special.logsumexp( loglikes - log_priors, axis=-1) - jnp.log(N)
        return self.aggregate_kernels(data, loglikes)

    def selection_cut(self, loglikes, N_eff):
        #print(N_eff)
        return jnp.where(N_eff > 4*self.N_events, loglikes, -jnp.nan_to_num(jnp.inf) * jnp.ones_like(loglikes))

    def event_cut(self, loglikes, N_eff):
        #print(N_eff, self.N_events, jnp.min(N_eff))
        return jnp.where(jnp.min(N_eff) > self.N_events, loglikes, -jnp.nan_to_num(jnp.inf) * jnp.ones_like(loglikes))

    def log_bayes_factors_event(self, params):
        if len(self.sampled_models) != 0:
            sampled_logweights_event = sum(self.log(model(self.event_data, params)) for model in self.sampled_models) - self.log(self.event_data["prior"])
        else:
            sampled_logweights_event = None

        ## Event analytic:
        if len(self.analytic_models) != 0:
            analytic_logweights_event = sum(self.log(model(self.event_data, params)) for model in self.analytic_models)
        else:
            analytic_logweights_event = None
        loglikes_event = self.event_expectation.log_bayes_factors(sampled_logweights=sampled_logweights_event, 
                                                                  analytic_logweights=analytic_logweights_event, 
                                                                  weights=self.event_data["weights"])
        if self.enforce_convergence:
            N_eff_events = self.event_expectation.N_eff(sampled_logweights=sampled_logweights_event, 
                                                        analytic_logweights=analytic_logweights_event, 
                                                        weights=self.event_data["weights"],
                                                        mu = jnp.exp(loglikes_event))

            return self.event_cut(loglikes_event, N_eff_events)


        return loglikes_event

    def log_bayes_factors_selection(self, params):
        data = self.selection_data.selection_data
        if len(self.sampled_models) != 0:
            sampled_logweights_selection = sum(self.log(model(data, params)) for model in self.selection_sampled_models) - self.log(data["prior"])
        else:
            sampled_logweights_selection = None

        if len(self.analytic_models) != 0:
            analytic_logweights_selection = sum(self.log(model(data, params)) for model in self.selection_analytic_models)
        else:
            analytic_logweights_selection = None

        loglikes_selection = self.selection_expectation.log_bayes_factors(sampled_logweights=sampled_logweights_selection, 
                                                                          analytic_logweights=analytic_logweights_selection, 
                                                                          weights=data["weights"])

        if self.enforce_convergence:
            N_eff_selection = self.selection_expectation.N_eff(sampled_logweights=sampled_logweights_selection, 
                                                                analytic_logweights=analytic_logweights_selection, 
                                                                weights=data["weights"],
                                                                mu = jnp.exp(loglikes_selection),
                                                                total_generated=self.selection_data.total_generated)
            return self.selection_cut(loglikes_selection, N_eff_selection)


        return loglikes_selection
    
    def logpdf(self, params):
        # Event Likelihoods
        loglikes_event = self.log_bayes_factors_event(params)
                
        # Selection Likelihood
        loglikes_selection = 0.0
        if self.selection_data:
            loglikes_selection += self.log_bayes_factors_selection(params)

        return loglikes_event.sum() - self.N_events * loglikes_selection

    @classmethod
    def from_file(cls, event_data_filename, selection_data_filename, sampled_models, analytic_models, 
                       SelectionClass=SelectionFunction, enforce_convergence=False, ignore_events=[], 
                       downsample=None, inflate_selection_spins=False, inflate_selection_spins_factor=4, downsample_selection=False,
                       selection_sampled_models=None, selection_analytic_models=None, kernel_fix_iterations=10):
        event_data, event_names = stack_nested_jax_arrays(load_hdf5_to_jax_dict(event_data_filename, ignore_events=ignore_events))
        #print(event_data.keys())

        if (len(sampled_models) != 0) and (downsample is not None):
            varname = sampled_models[0].var_names[0]
            #print(varname, len(event_data[varname].shape))
            E,K,N = event_data[varname].shape
            inds = np.random.randint(0,N, size=downsample)
            for col in event_data.keys():
                the_var_to_change = event_data[col]
                its_an_array = hasattr(the_var_to_change, 'shape')
                if its_an_array:
                    if len(the_var_to_change.shape) == 3:
                        event_data[col] = event_data[col][:,:,inds]

        redshift_model = [model for model in sampled_models if isinstance(model, Redshift)]
        redshift_model += [model for model in analytic_models if isinstance(model, Redshift)]
        if len(redshift_model) == 0:
            redshift_model = None
        else:
            redshift_model = redshift_model[0]

        if selection_data_filename is not None:
            selection_data = load_hdf5_to_jax_dict(selection_data_filename)
            selection_attributes = load_hdf5_attributes(selection_data_filename)
            if "selection" in selection_data.keys():
                selection_data = selection_data["selection"]

            if "selection" in selection_attributes.keys():
                selection_attributes = selection_attributes["selection"]

            if (len(sampled_models) != 0) and (downsample is not None) and (downsample_selection):
                varname = sampled_models[0].var_names[0]
                K,N = selection_data[varname].shape
                inds = np.random.randint(0,N, size=downsample)
                for col in selection_data.keys():
                    the_var_to_change = selection_data[col]
                    its_an_array = hasattr(the_var_to_change, 'shape')
                    if its_an_array:
                        if len(selection_data[col].shape) == 2:
                            selection_data[col] = selection_data[col][:,inds]


            if inflate_selection_spins:
                selection_data['chi_1_sigma_kernel'] *= inflate_selection_spins_factor
                selection_data['chi_2_sigma_kernel'] *= inflate_selection_spins_factor

            selection = SelectionClass(selection_data, 
                                       selection_attributes['analysis_time'],
                                       selection_attributes['total_generated'],
                                       redshift_model)
        else:
            selection = None

        return cls(sampled_models=sampled_models, analytic_models=analytic_models, selection_sampled_models=selection_sampled_models, 
                  selection_analytic_models=selection_analytic_models, event_data=event_data, 
                  selection_data=selection, enforce_convergence=enforce_convergence, event_names=event_names, kernel_fix_iterations=kernel_fix_iterations)



from ..utils import *
from ..generic import *
import jax.numpy as jnp
import jax

def _smoothing(masses, mmin, mmax, delta_m):
    shifted_mass = jnp.nan_to_num((masses - mmin) / delta_m, nan=0)
    shifted_mass = jnp.clip(shifted_mass, 1e-6, 1 - 1e-6)
    exponent = 1 / shifted_mass - 1 / (1 - shifted_mass)
    window = jax.scipy.special.expit(-exponent)
    return window*box(masses, mmin, mmax)

def smoothing(masses, mmin, mmax, delta_m):
    return jnp.where(delta_m > 0.0,  _smoothing(masses, mmin, mmax, delta_m), jnp.ones_like(masses))

def two_component_single(mass, alpha, lam, mmin, mmax, mpp, sigpp, gaussian_mass_maximum=100):
    p_pow = powerlaw(mass, alpha=-alpha, high=mmax, low=mmin)
    p_norm = truncnorm(mass, mu=mpp, sigma=sigpp, high=gaussian_mass_maximum, low=mmin)
    prob = (1 - lam) * p_pow + lam * p_norm
    return prob

class SmoothedTwoComponentPrimaryMassRatio(SampledPopulationModel, MassPopulationModel):
    def __init__(self, gaussian_mass_maximum=100, mmin_fixed=2, mmax_fixed=100, 
        primary_mass_name=None, mass_ratio_name=None, 
        normalization_shape=(1000,500), var_names=['mass_1', 'mass_ratio'],
        hyper_var_names=['alpha', 'beta', 'lam', 'mpp', 'sigpp', 'delta_m', 'mmin', 'mmax']):
        var_names[0] = primary_mass_name or var_names[0]
        var_names[1] = mass_ratio_name or var_names[1]
        self.hyper_var_names = hyper_var_names
        self.var_names = var_names
        self.gaussian_mass_maximum = gaussian_mass_maximum
        self.mmin_fixed = 2
        self.mmax_fixed = 100
        self.primary_mass_name = self.var_names[0]
        self.mass_ratio_name = self.var_names[1]
        self.m1s = jnp.linspace(mmin_fixed, mmax_fixed, normalization_shape[0])
        self.qs = jnp.linspace(mmin_fixed/mmax_fixed, 1, normalization_shape[1])
        self.m1s_grid, self.qs_grid = jnp.meshgrid(self.m1s, self.qs)
        self.mass_ratio_norm_clip_threshold = 1e-16
        #print("""Note: SmoothedTwoComponentPrimaryMassRatio is an unnormalized distribution. 
        #    Be wary when using these for infering merger rates. 
        #    In addition, this model might have a different primary mass marginal due to this lack of normalization in q""")

    def norm_p_q(self, data, beta, mmin, delta_m):
        """Calculate the mass ratio normalisation by linear interpolation"""
        p_q = powerlaw(self.qs_grid, beta, 1, mmin / self.m1s_grid)
        p_q *= smoothing(
            self.m1s_grid * self.qs_grid, mmin=mmin, mmax=self.m1s_grid, delta_m=delta_m
        )
        norms = jnp.where(
            jnp.array(delta_m) > 0,
            jax.scipy.integrate.trapezoid(jnp.nan_to_num(p_q), self.qs, axis=0),
            jnp.ones(self.m1s.shape),
        )

        result = jnp.interp(data[self.primary_mass_name], self.m1s, norms)
        return jnp.clip(result, self.mass_ratio_norm_clip_threshold)  # For preventing 1/0 at masses that are below mmin

    def norm_p_m1(self, delta_m, **kwargs):
        """Calculate the normalisation factor for the primary mass"""
        mmin = kwargs.get("mmin", self.mmin)
        p_m = self.__class__.primary_model(self.m1s, **kwargs)
        p_m *= smoothing(self.m1s, mmin=mmin, mmax=self.mmax, delta_m=delta_m)

        norm = jnp.where(jnp.array(delta_m) > 0, jax.scipy.integrate.trapezoid(p_m, self.m1s), 1)
        return norm

    def __call__(self, data, params):
        # Get params
        alpha = params[self.hyper_var_names[0]]
        lam = params[self.hyper_var_names[2]]
        mmin = params.get(self.hyper_var_names[6], self.mmin_fixed)
        mmax = params.get(self.hyper_var_names[7], self.mmax_fixed)
        beta = params[self.hyper_var_names[1]]
        mpp = params[self.hyper_var_names[3]]
        sigpp = params[self.hyper_var_names[4]]
        delta_m = params.get(self.hyper_var_names[5], 0)

        # Compute primary mass unnormalized distribution with smoothing
        p_m1 = two_component_single(data[self.primary_mass_name],  alpha, lam, mmin, mmax, mpp, sigpp, gaussian_mass_maximum=self.gaussian_mass_maximum)
        p_m1 *= smoothing(data[self.primary_mass_name], mmin, mmax, delta_m)

        p_m1_grid = two_component_single(self.m1s, alpha, lam, mmin, mmax, mpp, sigpp, gaussian_mass_maximum=self.gaussian_mass_maximum)
        p_m1_grid *= smoothing(self.m1s, mmin, mmax, delta_m)
        norm = jnp.where(delta_m > 0, jax.scipy.integrate.trapezoid(jnp.nan_to_num(p_m1_grid), self.m1s), 1)
        p_m1 /= norm

        # Compute mass_ratio unnormalized distribution with smoothing
        p_q = powerlaw(data[self.mass_ratio_name], beta, 1, mmin / data[self.primary_mass_name])
        p_q *= smoothing(
            data[self.primary_mass_name] * data[self.mass_ratio_name],
            mmin=mmin,
            mmax=data[self.primary_mass_name],
            delta_m=delta_m,
        )
        p_q /= self.norm_p_q(data=data, beta=beta, mmin=mmin, delta_m=delta_m)

        return jnp.nan_to_num(p_m1 * p_q, nan=0)



from dataclasses import dataclass, field
from ..models import Redshift, AbstractPopulationModel
import jax
import jax.numpy as jnp
from typing import Dict, List, Tuple, Union, Optional
import math
import numpy as np
import scipy
from ..utils import chunked_vmap
from tqdm import trange

@dataclass
class SelectionFunction:
	selection_data : Dict[str, jax.Array]
	analysis_time : float = 1
	total_generated : Optional[int] = None
	redshift_model : Optional[AbstractPopulationModel] = None
	total_detected :Optional[int] = None

	def __post_init__(self):
		self.total_detected = self.total_detected or math.prod(self.selection_data['prior'].shape)
		self.total_generated = self.total_generated or self.total_detected
		self.detection_ratio = self.total_detected/self.total_generated

	def surveyed_hypervolume(self, params):
		return self.redshift_model.normalization(params) / 1e9 * self.analysis_time

	def rate(self, likelihood, params, N_events=None):
		hyper_vol = self.surveyed_hypervolume(params)
		efficiency = jnp.exp(likelihood.total_event_bayes_factors(self.selection_data, params, N=self.total_generated))
		N_events = N_events or (likelihood.event_data['prior'].shape[0])
		return scipy.stats.gamma.rvs(a=N_events) / (efficiency * hyper_vol)

	def calculate_rate_for_hyperparameters_chunked(self, likelihood, params, N_events=None, chunk=100):
		N_events = N_events or (likelihood.N_events)
		arbitrary_variable_in_params = next(iter(params.keys()))
		param_shape = params[arbitrary_variable_in_params].shape
		in_axes_dict_shape = {variable : 0  for variable in params.keys()}
		prog_note_1 = "Calculating Rates : Selection Function ..."
		prog_note_2 = "Calculating Rates : Surveyed Hypervolume ..."
		efficiency_func = chunked_vmap(lambda params : jnp.exp(likelihood.total_event_bayes_factors(self.selection_data, params, N=self.total_generated)), in_axes=(in_axes_dict_shape,), chunk=chunk, progress_note=prog_note_1)
		hyper_vol_func = chunked_vmap(lambda params : self.surveyed_hypervolume(params), in_axes=(in_axes_dict_shape,), chunk=chunk, progress_note=prog_note_2)

		efficiency = efficiency_func(params)
		hyper_vol = hyper_vol_func(params)

		return scipy.stats.gamma.rvs(a=N_events, size=param_shape) / (efficiency * hyper_vol)

	def calculate_rate_for_hyperparameters(self, likelihood, params, N_events=None):
		arbitrary_variable_in_params = next(iter(params.keys()))
		N = params[arbitrary_variable_in_params].size

		list_of_params = [{key: params[key][i] for key in params.keys()} for i in range(N)]
		result = []
		for i in trange(N):
			result.append(self.rate(likelihood, list_of_params[i], N_events=N_events))

		return jnp.array(result)

	def calculate_N_eff_chunked(self, likelihood, params, chunk=100):
		N_eff_func = lambda params : likelihood.compute_selection_N_eff_only(params, N=self.total_generated)
		axes_dict_shape = {key:0 for key in params.keys()}
		N_eff_chunked_func = chunked_vmap(N_eff_func, in_axes=(axes_dict_shape,), chunk=100, progress_note="Calculating Selection Function N_effective ...")

		N_effs = N_eff_chunked_func(params)

		return N_effs







# -*- coding: utf-8 -*-

from setuptools import setup, find_packages


with open('README.md') as f:
    readme = f.read()

setup(
    name='gravpop',
    version='0.1.1',
    description="""Package to perform astrophysical population modelling using gravitational waves, 
                   specifically for modelling using Truncated Gaussian Mixtures or a hybrid scheme 
                   -large portions of this code are structured and computed very similar to gwpop""",
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    author='Asad Hussain',
    author_email='asadh@utexas.edu',
    url='https://github.com/potatoasad/gravpop',
    packages=find_packages(exclude=('tests', 'docs', 'dev'))
)



import jax
import jax.numpy as jnp
from ..utils import *
from .abstract import *
from ..sample import *

class TruncatedGaussian1D(SampledPopulationModel):
    r"""
    Truncated Gaussian Distribution. 

    Event Level Parameters:         :math:`x`
    Population Level Parameters:    :math:`\mu, \sigma` 

    .. math::
    
        P(x | \mu, \sigma) = \mathcal{N}_{[a,b]}))(x | \mu, \sigma) 
    """
    def __init__(self, a, b, var_names=['x'], hyper_var_names=['mu', 'sigma']):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.var_name = var_names[0]
        self.mu_name = hyper_var_names[0]
        self.sigma_name = hyper_var_names[1]
        
    def get_data(self, data, params):
        Xs          = data[self.var_name];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name]
        return Xs, mu, sigma
    
    def __call__(self, data, params):
        Xs, mu, sigma = self.get_data(data, params);
        #loglikes = jax.scipy.special.logsumexp( jnp.log(truncnorm(Xs, mu, sigma, low=a, high=b)) , axis=-1) - jnp.log(Xs.shape[-1])
        prob = truncnorm(Xs, mu, sigma, low=self.a, high=self.b)
        return prob


from jax import jit

@jit
def logcdf_inside(alpha, beta):
    return jax.scipy.special.logsumexp(
                jnp.stack([jax.scipy.stats.norm.logcdf(beta),
                           jax.scipy.stats.norm.logcdf(alpha)],
                          axis=-1),
                b=jnp.array([1,-1]),
                axis=-1
            )

@jit
def logcdf_right(alpha, beta):
    return jax.scipy.special.logsumexp(
                jnp.stack([jax.scipy.stats.norm.logcdf(beta),
                           jax.scipy.stats.norm.logcdf(alpha)],
                          axis=-1),
                b=jnp.array([1,-1]),
                axis=-1
            )

@jit
def logcdf_left(alpha, beta):
    return jax.scipy.special.logsumexp(
                jnp.stack([jax.scipy.stats.norm.logsf(alpha),
                           jax.scipy.stats.norm.logsf(beta)],
                          axis=-1),
                b=jnp.array([1,-1]),
                axis=-1
            )

@jit
def logC(mu, sigma, a, b):
    alpha = (a-mu)/sigma
    beta = (b-mu)/sigma
    return jnp.where((alpha > 2) & (beta > 2), logcdf_left(alpha, beta), # Way past negative
                    jnp.where( (alpha < -2) & (beta < -2), logcdf_right(alpha, beta), # Way past positive
                             logcdf_inside(alpha, beta) )) # Inside

@jit
def logC_deprecated(mu, sigma, a , b):
    return jax.scipy.special.logsumexp(
                jnp.stack([jax.scipy.stats.norm.logcdf(b, loc=mu, scale=sigma),
                           jax.scipy.stats.norm.logcdf(a, loc=mu, scale=sigma)],
                          axis=-1),
                b=jnp.array([1,-1]),
                axis=-1
            )

@jit
def transform_gaussians(x,dx,mu,sigma):
    #x = K.x; Δx = K.Δx;
    denom = dx**2 + sigma**2
    x_ = (dx**2 * mu + sigma**2 * x)/denom;
    sigma_ = jnp.sqrt((sigma**2 * dx**2)/denom);
    return x_, sigma_

@jit
def logweight(x,dx,mu,sigma,a,b):
    x_, sigma_ = transform_gaussians(x,dx,mu,sigma);
    
    logC_ = logC(x_, sigma_, a, b); #print(logC_)
    logCmu = logC(mu ,sigma , a, b);# print(logCmu)
    logCx = logC(x ,dx , a, b); #print(logCx)
    logA = logC_ - logCmu - logCx
    return logA

@jit
def logweight_fixed_data_limits(x,dx,mu,sigma,a,b, a_data, b_data):
    x_, sigma_ = transform_gaussians(x,dx,mu,sigma);
    
    logC_ = logC(x_, sigma_, a, b); #print(logC_)
    logCmu = logC(mu ,sigma , a, b);# print(logCmu)
    logCx = logC(x ,dx , a_data, b_data); #print(logCx)
    logA = logC_ - logCmu - logCx
    return logA

@jit
def loglikelihood_kernel1d(x, dx, mu, sigma, a, b):
    logA = logweight(x,dx,mu,sigma,a,b)
    logB = jax.scipy.stats.norm.logpdf(x,loc=mu, scale=jnp.sqrt(dx**2 + sigma**2))     #normlogpdf(x,jnp.sqrt(Δx^2 + σ^2), μ)
    return logA + logB

@jit
def loglikelihood_kernel1d_fixed_data_limits(x, dx, mu, sigma, a, b, a_data, b_data):
    logA = logweight_fixed_data_limits(x,dx,mu,sigma,a,b, a_data, b_data)
    logB = jax.scipy.stats.norm.logpdf(x,loc=mu, scale=jnp.sqrt(dx**2 + sigma**2))     #normlogpdf(x,jnp.sqrt(Δx^2 + σ^2), μ)
    return logA + logB


class TruncatedGaussian1DAnalytic(AnalyticPopulationModel):
    r"""
    Truncated Gaussian Distribution. Evaluates an analytical expression for the population likelihood of a truncated
    gaussian distribution evaluated over data distributed as a truncated gaussian as well. Useful when our data is represented as a truncated gaussian mixture. 
    This class evaluated over the components of such a mixture, would allow us to evaluate the population likelihood of arbitrary data distributions represented using truncated gaussian mixtures. 

    Event Level Parameters:         :math:`x \sim \mathcal{N}_{[a,b]}(x_0, \Delta x)` 
    Population Level Parameters:    :math:`\mu, \sigma` 

    .. math::
    
        P(x | \mu, \sigma) = \mathcal{N}_{[a,b]}))(x | \mu, \sigma) \textrm{ where } x \sim \mathcal{N}_{[a,b]}(x_0, \Delta x)
    """
    def __init__(self, a, b, var_names=['x'], hyper_var_names=['mu', 'sigma']):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.var_name = var_names[0]
        self.mu_name = hyper_var_names[0]
        self.sigma_name = hyper_var_names[1]

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}
        
    def get_data(self, data, params):
        X_locations = data[self.var_name + '_mu_kernel'];
        X_scales    = data[self.var_name + '_sigma_kernel'];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name]
        return X_locations, X_scales, mu, sigma

    def evaluate(self, data, params):
        """
        Evaluates the value of the *integrand* and not the integral
        """
        Xs          = data[self.var_name];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name];
        #loglikes = jax.scipy.special.logsumexp( jnp.log(truncnorm(Xs, mu, sigma, low=a, high=b)) , axis=-1) - jnp.log(Xs.shape[-1])
        prob = truncnorm(Xs, mu, sigma, low=self.a, high=self.b)
        return prob
        
    def __call__(self, data, params):
        X_locations, X_scales, mu, sigma = self.get_data(data, params);
        loglikes = loglikelihood_kernel1d(X_locations, X_scales, mu, sigma, self.a, self.b)
        return jnp.exp(loglikes)

    def sample(self, df_hyper_samples, oversample=1):
        return ppd_truncCorrelatedanalytic(self, df_hyper_samples, oversample=oversample)


class TruncatedGaussian1DAnalyticVariedLimits(AnalyticPopulationModel):
    r"""
    Truncated Gaussian Distribution. Evaluates an analytical expression for the population likelihood of a truncated
    gaussian distribution evaluated over data distributed as a truncated gaussian as well. Useful when our data is represented as a truncated gaussian mixture. 
    This class evaluated over the components of such a mixture, would allow us to evaluate the population likelihood of arbitrary data distributions represented using truncated gaussian mixtures. 

    Event Level Parameters:         :math:`x \sim \mathcal{N}_{[a,b]}(x_0, \Delta x)` 
    Population Level Parameters:    :math:`\mu, \sigma` 

    .. math::
    
        P(x | \mu, \sigma) = \mathcal{N}_{[a,b]}))(x | \mu, \sigma) \textrm{ where } x \sim \mathcal{N}_{[a,b]}(x_0, \Delta x)
    """
    def __init__(self, a, b, var_names=['x'], hyper_var_names=['mu', 'sigma', 'x_min', 'x_max']):
        self.a = a
        self.b = b
        self.var_names = var_names
        self.hyper_var_names = hyper_var_names
        self.var_name = var_names[0]
        self.mu_name = hyper_var_names[0]
        self.sigma_name = hyper_var_names[1]
        self.x_min = hyper_var_names[2];
        self.x_max = hyper_var_names[3];

    @property
    def limits(self):
        return {var : [self.a, self.b] for i,var in enumerate(self.var_names)}
        
    def get_data(self, data, params):
        X_locations = data[self.var_name + '_mu_kernel'];
        X_scales    = data[self.var_name + '_sigma_kernel'];
        mu          = params[self.mu_name]
        sigma       = params[self.sigma_name]
        x_min       = params[self.x_min]
        x_max       = params[self.x_max]
        return X_locations, X_scales, mu, sigma, x_min, x_max

    def evaluate(self, data, params):
        """
        Evaluates the value of the *integrand* and not the integral
        """
        Xs          = data[self.var_name];
        mu          = params[self.mu_name];
        sigma       = params[self.sigma_name];
        x_min       = params[self.x_min];
        x_max       = params[self.x_max];
        #loglikes = jax.scipy.special.logsumexp( jnp.log(truncnorm(Xs, mu, sigma, low=a, high=b)) , axis=-1) - jnp.log(Xs.shape[-1])
        prob = truncnorm(Xs, mu, sigma, low=x_min, high=x_max)
        return prob
        
    def __call__(self, data, params):
        X_locations, X_scales, mu, sigma, x_min, x_max = self.get_data(data, params);
        loglikes = loglikelihood_kernel1d_fixed_data_limits(X_locations, X_scales, mu, sigma, x_min, x_max, self.a, self.b)
        return jnp.exp(loglikes)

    def sample(self, df_hyper_samples, oversample=1):
        return ppd_truncCorrelatedanalytic(self, df_hyper_samples, oversample=oversample)









import jax.numpy as jnp
import numpy as np
from dataclasses import dataclass
import numpyro.distributions as dist
from typing import Dict
import jax
from gravpop import * 
import matplotlib.pyplot as plt
import scipy
import pytest

def test_covariance():
	sig_1, sig_2, rho = np.random.rand()*3, np.random.rand()*3, (np.random.rand()*2 - 1)
	Sigma_mat = np.array([[sig_1**2, sig_1*sig_2*rho],[sig_1*sig_2*rho, sig_2**2]])
	Sigma_test = CovarianceMatrix2D(sig_1, sig_2, rho)
	assert np.abs(np.linalg.det(np.linalg.inv(Sigma_mat) - Sigma_test.inv().value))


import jax.numpy as jnp
import numpy as np
from dataclasses import dataclass
import numpyro.distributions as dist
from typing import Dict
import jax
from gravpop import * 
import matplotlib.pyplot as plt
import scipy
import pytest


def test_likelihood_sampled():
	TG = Gaussian1D(var_name='x', mu_name='mu', sigma_name='sigma')
	mu = 0.0
	sigma = 0.2
	likelihood_sigma = 1.0

	# Number of Events, Kernels & Samples
	E, K, N = 10, 1, 1000  

	events = scipy.stats.norm.rvs(loc=mu, scale=sigma, size=E)
	x_data = jnp.stack([scipy.stats.norm.rvs(loc=x, scale=likelihood_sigma, size=N) for x in events]).reshape(E,K,N)


	event_data = {'x': x_data, # 10 events, 1 kernels, 1000 points each
				  'weights': jnp.ones((E,K))/K}

	HL = HybridPopulationLikelihood(sampled_models=[TG], event_data=event_data, analytic_models=[])



	Samp = Sampler(priors = {'mu' : dist.Normal(0,3), 'sigma' : dist.Uniform(0,4)}, 
				   latex_symbols = {'mu' : r'$\mu$','sigma' : r'$\sigma$'} , 
				   likelihood=HL, num_samples=2000, num_warmup=1000)

	Samp.sample();

	fig = Samp.corner(truth=np.array([mu,sigma]));
	fig.savefig("./test/test_likelihood_plot.png")


def test_constructed_full_likelihood():
	filename = 'test/testdata/test_event_data.h5'
	selection_filename = 'test/testdata/test_selection_function.h5'

	SM = SmoothedTwoComponentPrimaryMassRatio(primary_mass_name="mass_1_source")
	R = PowerLawRedshift(z_max=1.9)

	HL = PopulationLikelihood.from_file(
	            event_data_filename = filename,
	            selection_data_filename = selection_filename,
	            models = [SM,R]
	            )

	Lambda_0 = dict(alpha = 3.5,lam = 0.04,mmin = 5,mmax = 96,beta = 1.1,mpp = 35,sigpp = 4,delta_m = 3,lamb = 2.9)
	HL.logpdf(Lambda_0)

	Lambda_0 = dict(alpha = [3.5],lam = [0.04],mmin = [5],mmax = [96],beta = [1.1],mpp = [35],sigpp = [4],delta_m = [3],lamb = [2.9])
	Lambda_0 = {key: jnp.array(value) for key, value in Lambda_0.items()}
	HL.selection_data.calculate_N_eff_chunked(HL, Lambda_0, chunk=100)
		

@pytest.mark.slow
def test_posterior_pulling_data():
	filename = 'test/testdata/test_event_data.h5'
	selection_filename = 'test/testdata/test_selection_function.h5'
	posterior_file = "/Users/asadh/Documents/GitHub/gravpop_pipe/dev/samples_output.csv"

	SM = SmoothedTwoComponentPrimaryMassRatio(primary_mass_name="mass_1_source")
	R = PowerLawRedshift(z_max=1.9)

	HP = HyperPosterior.from_file(posterior_file, filename, selection_filename, models=[SM,R])
	#HP.calculate_rates()
	HP.N_effective_cuts()
	HP.posterior_with_cuts.to_csv("/Users/asadh/Documents/GitHub/gravpop_pipe/dev/samples_output_w_cuts.csv",index=False)








import pytest

from gravpop import *
import gravpop
import gwpopulation
from .testing_utils import *

from matplotlib.gridspec import GridSpec
import matplotlib.pyplot as plt


# Define your test class
class TestCompareToGwpopulation:
    # Define setup method if needed
    def setup_method(self, method):
        ## GLOBAL POINTS
        self.evaluation_point = GLOBAL_DICTIONARY_OF_NICE_VALUES
        #self.evaluation_point = GLOBAL_DICTIONARY_OF_EVIL_VALUES
        self.prior = PRIOR_DICT 

        #### Mass Models
        self.mass_params =  ['alpha','beta','mmin','mmax','mpp','sigpp','lam','delta_m']
        self.mass_point = {k:v for k,v in self.evaluation_point.items() if k in self.mass_params}
        self.mass_grid = Grid([Grid1D(name='mass_1', minimum=2, maximum=100, N=200), 
                               Grid1D(name='mass_ratio', minimum=0, maximum=1  , N=200)]).data

        self.gwpop_mass_model = gwpopulation.models.mass.SinglePeakSmoothedMassDistribution()
        self.gravpop_mass_model = gravpop.SmoothedTwoComponentPrimaryMassRatio(primary_mass_name='mass_1')

        #### Redshift Model
        self.redshift_params =  ['lamb']
        self.redshift_point = {k:v for k,v in self.evaluation_point.items() if k in self.redshift_params}

        self.evaluation_point_redshift = GLOBAL_DICTIONARY_OF_NICE_VALUES
        self.redshift_grid = Grid1D(name='redshift', minimum=0, maximum=1.9, N=200).data

        self.gwpop_redshift_model = gwpopulation.models.redshift.PowerLawRedshift(z_max=1.9)
        self.gravpop_redshift_model = gravpop.PowerLawRedshift(z_max=1.9)
    
    # Define teardown method if needed
    def teardown_method(self, method):
        pass
    
    # Write your test methods
    def test_mass_model_matches_gwpop(self):
        Z_gwpop = self.gwpop_mass_model(jax_to_numpy(self.mass_grid), **self.mass_point)
        Z_gravpop = self.gravpop_mass_model(self.mass_grid, self.evaluation_point)

        #fig, axes = plt.subplots(ncols=2, figsize=(10,5))
        from matplotlib.gridspec import GridSpec
        import matplotlib.pyplot as plt

        fig = plt.figure(layout="constrained")
        gs = GridSpec(2, 2, figure=fig)
        ax1 = fig.add_subplot(gs[0, 0])
        ax2 = fig.add_subplot(gs[0, 1])
        ax3 = fig.add_subplot(gs[1, :])
        ax1.contourf(self.mass_grid['mass_1'], self.mass_grid['mass_ratio'], jnp.log(Z_gravpop))
        ax2.contourf(self.mass_grid['mass_1'], self.mass_grid['mass_ratio'], jnp.log(Z_gwpop))
        ax1.set_xlabel(r"$m_1$"); ax2.set_xlabel(r"$m_1$");
        ax1.set_ylabel(r"$q$"); ax2.set_ylabel(r"$q$");
        ax1.set_title(r"$\log P_{gravpop}$")
        ax2.set_title(r"$\log P_{gwpopulation}$")
        ax3.hist(jnp.abs(Z_gravpop/Z_gwpop).flatten(), bins=100);
        ax3.set_title(r"Histogram of $P_{gravpop}/P_{gwpop}$")
        plt.suptitle("Log probability comparison for mass models")
        fig.savefig("./test/images/mass_model_comparison.png")
        assert not jnp.any(jnp.isnan(Z_gravpop)) # TESTING NO NANs

    def test_mass_model_gradient(self):
        derivatives = model_gradient(self.gravpop_mass_model, self.mass_grid, self.mass_point)

        import matplotlib.pyplot as plt
        from matplotlib.gridspec import GridSpec

        keys = list(derivatives.keys())
        n = len(keys)
        fig = plt.figure(layout="constrained", figsize=(10, 5*n))
        gs = GridSpec(n, 2, figure=fig)

        axes = []
        for i in range(n):
            row_axes = []
            for j in range(2):
                ax = fig.add_subplot(gs[i, j])
                row_axes.append(ax)
            axes.append(row_axes)

        for i, ax_row in enumerate(axes):
            ax = ax_row[0]
            M1, Q = self.mass_grid["mass_1"], self.mass_grid["mass_ratio"]
            contour = ax.contourf(M1,Q,derivatives[keys[i]])
            cbar = plt.colorbar(contour, ax=ax)
                
            # Add axis labels and title
            ax.set_xlabel(r"$m_1$")
            ax.set_ylabel(r"$q$")
            ax.set_title(f"{keys[i]} \n Gradient of model probability w.r.t to {keys[i]}")
            
            ax = ax_row[1]
            null_values_mask = jnp.isnan(derivatives[keys[i]])
            ax.imshow(null_values_mask, extent=(M1.min(), M1.max(),
                                                 Q.min(), Q.max()),
                      cmap='binary', alpha=0.5, origin='lower', aspect='auto')
            
           #ax.set_aspect('equal')
                
            # Add axis labels and title
            ax.set_xlabel(r"$m_1$")
            ax.set_ylabel(r"$q$")
            ax.set_title(f"{keys[i]} \n Locations where model probability gradient was null")
            
            
        total_null_points = jnp.sum(jnp.array([jnp.sum(jnp.isnan(derivatives[x])) for x in derivatives.keys()]))
        plt.suptitle(f"A total of {total_null_points} points had null gradients")
        fig.savefig("./test/images/mass_model_gradient_checks.png")
        assert not jnp.any(jnp.array([jnp.any(jnp.isnan(derivatives[x])) for x in derivatives.keys()]))

    def test_redshift_model_matches_gwpop(self):
        Z_gwpop = self.gwpop_redshift_model(jax_to_numpy(self.redshift_grid), **self.redshift_point)
        Z_gravpop = self.gravpop_redshift_model(self.redshift_grid, self.evaluation_point)

        fig = plt.figure(layout="constrained")
        gs = GridSpec(1, 1, figure=fig)
        ax1 = fig.add_subplot(gs[0, 0])

        lambs = np.linspace(-6,6,20)
        lambs = [{'lamb' : l} for l in lambs]
        zs = self.redshift_grid['redshift']

        fig,ax = plt.subplots(1)
        for lamb in lambs:
            Z_gwpop_inner = self.gwpop_redshift_model(jax_to_numpy(self.redshift_grid), **lamb)
            Z_gravpop_inner = self.gravpop_redshift_model(self.redshift_grid, lamb)
            dummy_1 = ax.plot(zs, Z_gravpop_inner, color="r", alpha=0.5, label="gravpop")
            dummy_2 = ax.plot(zs, Z_gwpop_inner, color="b", alpha=0.5, label="gwpopulation")

        ax.set_xlabel(r"$z$"); ax.set_ylabel(fr"$P(z | \kappa)$")
        plt.legend(handles=[dummy_1[0], dummy_2[0]])
        plt.title("Probability comparison for redshift models")
        fig.savefig("./test/images/redshift_model_comparison.png")
        assert not jnp.any(jnp.isnan(Z_gravpop)) # TESTING NO NANs

    def test_redshift_model_gradient(self):
        derivatives = model_gradient(self.gravpop_redshift_model, self.redshift_grid, self.redshift_point)
        fig,axes = plt.subplots(nrows=2, figsize=(5,10))

        ax = axes[0]
        zs = self.redshift_grid['redshift']
        ax.plot(zs, derivatives['lamb'], color="r", alpha=0.5)
        ax.set_xlabel(r"$z$"); ax.set_ylabel(r"$\frac{\partial P(z | \kappa)}{\partial \kappa}$")
        ax.set_title(r"$\frac{\partial P(z | \kappa)}{\partial \kappa}$")

        ax = axes[1]
        zs = self.redshift_grid['redshift']
        ax.scatter(zs, jnp.isnan(derivatives['lamb']), color="r", alpha=0.5, marker='o', facecolor='none', edgecolor='b')
        ax.set_yticks([0, 1])
        ax.set_yticklabels(["Not Null", "Null"])
        ax.set_xlabel(r"$z$"); ax.set_ylabel(r"$\frac{\partial P(z | \kappa)}{\partial \kappa}$")
        ax.set_title("Check if gradient is null")

        fig.savefig("./test/images/redshift_model_gradient_checks.png")
        assert not jnp.any(jnp.array([jnp.any(jnp.isnan(derivatives[x])) for x in derivatives.keys()]))



# Define your test class
class TestMassRatioAtNegativeBeta:
    # Define setup method if needed
    def setup_method(self, method):
        ## GLOBAL POINTS
        self.evaluation_point = GLOBAL_DICTIONARY_OF_NICE_VALUES
        self.evaluation_point['beta'] = -1.0
        #self.evaluation_point = GLOBAL_DICTIONARY_OF_EVIL_VALUES
        self.prior = PRIOR_DICT 

        #### Mass Models
        self.mass_params =  ['alpha','beta','mmin','mmax','mpp','sigpp','lam','delta_m']
        self.mass_point = {k:v for k,v in self.evaluation_point.items() if k in self.mass_params}
        self.mass_grid = Grid([Grid1D(name='mass_1', minimum=2, maximum=100, N=200), 
                               Grid1D(name='mass_ratio', minimum=0, maximum=1  , N=200)]).data

        self.gwpop_mass_model = gwpopulation.models.mass.SinglePeakSmoothedMassDistribution()
        self.gravpop_mass_model = gravpop.SmoothedTwoComponentPrimaryMassRatio(primary_mass_name='mass_1')

        #### Redshift Model
        self.redshift_params =  ['lamb']
        self.redshift_point = {k:v for k,v in self.evaluation_point.items() if k in self.redshift_params}

        self.evaluation_point_redshift = GLOBAL_DICTIONARY_OF_NICE_VALUES
        self.redshift_grid = Grid1D(name='redshift', minimum=0, maximum=1.9, N=200).data

        self.gwpop_redshift_model = gwpopulation.models.redshift.PowerLawRedshift(z_max=1.9)
        self.gravpop_redshift_model = gravpop.PowerLawRedshift(z_max=1.9)
    
    # Define teardown method if needed
    def teardown_method(self, method):
        pass
    
    # Write your test methods
    def test_mass_model_matches_gwpop(self):
        Z_gwpop = self.gwpop_mass_model(jax_to_numpy(self.mass_grid), **self.mass_point)
        Z_gravpop = self.gravpop_mass_model(self.mass_grid, self.evaluation_point)

        #fig, axes = plt.subplots(ncols=2, figsize=(10,5))
        from matplotlib.gridspec import GridSpec
        import matplotlib.pyplot as plt

        fig = plt.figure(layout="constrained")
        gs = GridSpec(2, 2, figure=fig)
        ax1 = fig.add_subplot(gs[0, 0])
        ax2 = fig.add_subplot(gs[0, 1])
        ax3 = fig.add_subplot(gs[1, :])
        ax1.contourf(self.mass_grid['mass_1'], self.mass_grid['mass_ratio'], jnp.log(Z_gravpop))
        ax2.contourf(self.mass_grid['mass_1'], self.mass_grid['mass_ratio'], jnp.log(Z_gwpop))
        ax1.set_xlabel(r"$m_1$"); ax2.set_xlabel(r"$m_1$");
        ax1.set_ylabel(r"$q$"); ax2.set_ylabel(r"$q$");
        ax1.set_title(r"$\log P_{gravpop}$")
        ax2.set_title(r"$\log P_{gwpopulation}$")
        ax3.hist(jnp.abs(Z_gravpop/Z_gwpop).flatten(), bins=100);
        ax3.set_title(r"Histogram of $P_{gravpop}/P_{gwpop}$")
        plt.suptitle("Log probability comparison for mass models")
        fig.savefig("./test/images/mass_model_comparison_beta_negative.png")
        assert not jnp.any(jnp.isnan(Z_gravpop)) # TESTING NO NANs

    def test_mass_model_gradient(self):
        derivatives = model_gradient(self.gravpop_mass_model, self.mass_grid, self.mass_point)

        import matplotlib.pyplot as plt
        from matplotlib.gridspec import GridSpec

        keys = list(derivatives.keys())
        n = len(keys)
        fig = plt.figure(layout="constrained", figsize=(10, 5*n))
        gs = GridSpec(n, 2, figure=fig)

        axes = []
        for i in range(n):
            row_axes = []
            for j in range(2):
                ax = fig.add_subplot(gs[i, j])
                row_axes.append(ax)
            axes.append(row_axes)

        for i, ax_row in enumerate(axes):
            ax = ax_row[0]
            M1, Q = self.mass_grid["mass_1"], self.mass_grid["mass_ratio"]
            contour = ax.contourf(M1,Q,derivatives[keys[i]])
            cbar = plt.colorbar(contour, ax=ax)
                
            # Add axis labels and title
            ax.set_xlabel(r"$m_1$")
            ax.set_ylabel(r"$q$")
            ax.set_title(f"{keys[i]} \n Gradient of model probability w.r.t to {keys[i]}")
            
            ax = ax_row[1]
            null_values_mask = jnp.isnan(derivatives[keys[i]])
            ax.imshow(null_values_mask, extent=(M1.min(), M1.max(),
                                                 Q.min(), Q.max()),
                      cmap='binary', alpha=0.5, origin='lower', aspect='auto')
            
           #ax.set_aspect('equal')
                
            # Add axis labels and title
            ax.set_xlabel(r"$m_1$")
            ax.set_ylabel(r"$q$")
            ax.set_title(f"{keys[i]} \n Locations where model probability gradient was null")
            
            
        total_null_points = jnp.sum(jnp.array([jnp.sum(jnp.isnan(derivatives[x])) for x in derivatives.keys()]))
        plt.suptitle(f"A total of {total_null_points} points had null gradients")
        fig.savefig("./test/images/mass_model_gradient_checks_beta_negative.png")
        assert not jnp.any(jnp.array([jnp.any(jnp.isnan(derivatives[x])) for x in derivatives.keys()]))
